# Copyright (C) 2022 -- svg2mod developers < GitHub.com / svg2mod >

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

'''
Tools to convert data from Svg2ModImport to
the file information  used in kicad module files.
'''


import copy
import datetime
import io
import json
import os
import re
import time
from abc import ABC, abstractmethod

from svg2mod import svg
from svg2mod.coloredlogger import logger, unfiltered_logger
from svg2mod.importer import Svg2ModImport
from svg2mod.svg2mod import PolygonSegment

#----------------------------------------------------------------------------

DEFAULT_DPI = 96 # 96 as of Inkscape 0.92
MINIMUM_SIZE = 1e-5 # Minimum size kicad will render

#----------------------------------------------------------------------------

class Svg2ModExport(ABC):
    ''' An abstract class to provide functionality
    to write to kicad module file.
    The abstract methods are the file type specific
    example: pretty, legacy
    '''

    #------------------------------------------------------------------------

    @property
    @abstractmethod
    def layer_map(self ):
        ''' This should be overwritten by a dictionary object of layer maps '''
        pass

    @abstractmethod
    def _get_layer_name( self, item_name, name, front ):pass

    @abstractmethod
    def _write_library_intro( self, cmdline ): pass

    @abstractmethod
    def _get_module_name( self, front = None ): pass

    @abstractmethod
    def _write_module_header( self, label_size, label_pen, reference_y, value_y, front,): pass

    @abstractmethod
    def _write_modules( self ): pass

    @abstractmethod
    def _write_module_footer( self, front ):pass

    @abstractmethod
    def _write_polygon_header( self, points, layer ):pass

    @abstractmethod
    def _write_polygon_footer( self, layer, stroke_width, fill=True):pass

    @abstractmethod
    def _write_polygon_point( self, point ):pass

    @abstractmethod
    def _write_polygon_segment( self, p, q, layer, stroke_width ):pass

    @abstractmethod
    def _write_thru_hole( self, circle, layer ):pass

    #------------------------------------------------------------------------

    @staticmethod
    def _convert_decimal_to_mm( decimal ):
        return float( decimal ) * 0.00254


    #------------------------------------------------------------------------

    @staticmethod
    def _convert_mm_to_decimal( mm ):
        return int( round( mm * 393.700787 ) )


    #------------------------------------------------------------------------

    def _get_fill_stroke( self, item ):

        s = item.style

        fill = False if not s.get('fill') or s["fill"] == "none" else True
        fill = fill if not s.get('fill-opacity') or float(s['fill-opacity']) != 0 else False

        stroke = False if not s.get('stroke') or s["stroke"] == "none" else True
        stroke = stroke if not s.get('stroke-opacity') or float(s['stroke-opacity']) != 0 else False

        stroke_width = s["stroke-width"] * self.scale_factor if s.get('stroke-width') else MINIMUM_SIZE

        if stroke_width is None:
            stroke_width = 0

        # This should display something.
        if not self.imported.ignore_hidden and not fill and not stroke:
            stroke = True
            stroke_width = stroke_width if stroke_width else MINIMUM_SIZE

        # There should be no stroke_width if no stroke
        return fill, stroke, stroke_width if stroke else 0


    #------------------------------------------------------------------------

    def __init__(
        self,
        svg2mod_import = Svg2ModImport(),
        file_name = None,
        center = False,
        scale_factor = 1.0,
        precision = 20.0,
        use_mm = True,
        dpi = DEFAULT_DPI,
        pads = False,
    ):
        if use_mm:
            # 25.4 mm/in;
            scale_factor *= 25.4 / float(dpi)
            use_mm = True
        else:
            # PCBNew uses decimal (10K DPI);
            scale_factor *= 10000.0 / float(dpi)

        self.imported = svg2mod_import
        self.file_name = file_name
        self.center = center
        self.scale_factor = scale_factor
        self.precision = precision
        self.use_mm = use_mm
        self.dpi = dpi
        self.convert_pads = pads

        # Local instance variables
        self.translation = None
        self.layers = {}
        self.output_file = None
        self.raw_file_data = None


    #------------------------------------------------------------------------

    def add_svg_element(self, elem : svg.Transformable, layer="F.SilkS"):
        ''' This can be used to add a svg element
        to a specific layer.
        If the importer doesn't have a svg element
        it will also create an empty Svg object.
        '''
        grp = svg.Group()
        grp.name = layer
        grp.items.append(elem)
        try:
            self.imported.svg.items.append(grp)
        except AttributeError:
            self.imported.svg = svg.Svg()
            self.imported.svg.items.append(grp)

    #------------------------------------------------------------------------


    def _calculate_translation( self ):

        min_point, max_point = self.imported.svg.bbox()

        if self.center:
            # Center the drawing:
            adjust_x = min_point.x + ( max_point.x - min_point.x ) / 2.0
            adjust_y = min_point.y + ( max_point.y - min_point.y ) / 2.0

            self.translation = svg.Point(
                0.0 - adjust_x,
                0.0 - adjust_y,
            )

        else:
            self.translation = svg.Point(
                0.0,
                0.0,
            )

    #------------------------------------------------------------------------

    def _prune( self, items = None ):
        '''Find and keep only the layers of interest.'''

        empty_group_exception = items is None

        if items is None:

            self.layers = {}
            for name in self.layer_map.keys():
                self.layers[ name ] = []


            items = self.imported.svg.items
            self.imported.svg.items = []

        kept_layers = {}

        for item in items:

            if not isinstance( item, svg.Group ):
                continue
            i_name = item.name.split(":", 1)

            for name in self.layers.keys():
                # if name == i_name[0] and i_name[0] != "":
                if re.match( '^{}$'.format(name), i_name[0]):
                    # Don't add empty groups to the list of valid items
                    if isinstance(item, svg.Group) and not item.items:
                        break

                    if kept_layers.get(i_name[0]):
                        kept_layers[i_name[0]].append(item.name)
                    else:
                        kept_layers[i_name[0]] = [item.name]

                    self.imported.svg.items.append( item )
                    self.layers[name].append((i_name, item))
                    break
            else:
                self._prune( item.items )

        for kept in sorted(kept_layers.keys()):
            unfiltered_logger.info( "Found SVG layer: {}".format( kept ) )
            logger.debug( "  Detailed names: [{}]".format( ", ".join(kept_layers[kept]) ) )

        # There are no elements to write so don't write
        if empty_group_exception:
            for name in self.layers:
                if self.layers[name]:
                    break
            else:
                raise Exception("Not writing empty file. No valid items found.")

    #------------------------------------------------------------------------

    def _write_items( self, items, layer, flip = False ):

        for item in items:

            if isinstance( item, svg.Group ):
                self._write_items( item.items, layer, flip )
                continue

            if re.match(r"^Drill\.\w+", str(layer)):
                if isinstance(item, svg.Circle):
                    self._write_thru_hole(item, layer)
                else:
                    logger.warning( "Non Circle SVG element in drill layer: {}".format(item.__class__.__name__))

            elif isinstance( item, (svg.Path, svg.Ellipse, svg.Rect, svg.Text)):

                segments = [
                    PolygonSegment( segment )
                    for segment in item.segments(
                        precision = self.precision
                    )
                ]

                fill, stroke, stroke_width = self._get_fill_stroke( item )
                if layer == "Edge.Cuts":
                    fill = False
                    stroke = True
                    stroke_width = MINIMUM_SIZE if stroke_width < MINIMUM_SIZE else stroke_width


                fill = (True if re.match("^Keepout", str(layer)) else fill)
                stroke_width = (0.508 if re.match("^Keepout", str(layer)) else stroke_width)

                for segment in segments:
                    segment.process( self, flip, fill )

                if len( segments ) > 1:
                    # Sort segments in order of size
                    segments.sort(key=lambda v: svg.Segment(v.bbox[0], v.bbox[1]).length(), reverse=True)

                    # Write all segments
                    while len(segments) > 0:
                        inlinable = [segments[0]]

                        # Search to see if any paths are contained in the current shape
                        for seg in segments[1:]:
                            # Contained in parent shape
                            if fill and not inlinable[0].are_distinct(seg):
                                append = True
                                if len(inlinable) > 1:
                                    for hole in inlinable[1:]:
                                        # Contained in a hole. It is separate
                                        if not hole.are_distinct(seg):
                                            append = False
                                            break
                                if append: inlinable.append(seg)
                        for poly in inlinable:
                            segments.pop(segments.index(poly))
                        if len(inlinable) > 1:
                            points = inlinable[ 0 ].inline( inlinable[ 1 : ] )
                        elif len(inlinable) > 0:
                            points = inlinable[ 0 ].points

                        logger.info( "  Writing {} with {} points".format(item.__class__.__name__, len( points ) ))

                        self._write_polygon(
                            points, layer, fill, stroke, stroke_width
                        )
                    continue

                if len( segments ) > 0:
                    points = segments[ 0 ].points

                if len ( segments ) == 1:

                    logger.info( "  Writing {} with {} points".format(item.__class__.__name__, len( points ) ))

                    self._write_polygon(
                        points, layer, fill, stroke, stroke_width
                    )
                else:
                    logger.info( "  Skipping {} with 0 points".format(item.__class__.__name__))

            else:
                logger.warning( "Unsupported SVG element: {}".format(item.__class__.__name__))


    #------------------------------------------------------------------------

    def _write_module( self, front ):

        module_name = self._get_module_name( front )

        min_point, max_point = self.imported.svg.bbox()
        min_point = self.transform_point( min_point, flip = False )
        max_point = self.transform_point( max_point, flip = False )

        label_offset = 1200
        label_size = 600
        label_pen = 120

        if self.use_mm:
            label_size = self._convert_decimal_to_mm( label_size )
            label_pen = self._convert_decimal_to_mm( label_pen )
            reference_y = min_point.y - self._convert_decimal_to_mm( label_offset )
            value_y = max_point.y + self._convert_decimal_to_mm( label_offset )
        else:
            reference_y = min_point.y - label_offset
            value_y = max_point.y + label_offset

        self._write_module_header(
            label_size, label_pen,
            reference_y, value_y,
            front,
        )

        for name, groups in self.layers.items():
            for i_name, group in groups:

                if group is None: continue

                layer = self._get_layer_name( i_name, name, front )

                self._write_items( group.items, layer, not front )

        self._write_module_footer( front )


    #------------------------------------------------------------------------

    def _write_polygon( self, points, layer, fill, stroke, stroke_width ):

        if fill and len(points) > 2:
            self._write_polygon_filled(
                points, layer, stroke_width
            )
            return

        # Polygons with a fill and stroke are drawn with the filled polygon above
        if stroke:
            if len(points) == 1:
                points.append(copy.copy(points[0]))

            self._write_polygon_outline(
                points, layer, stroke_width
            )
            return

        if len(points) < 3:
            logger.debug("  Not writing non-polygon with no stroke.")
        else:
            logger.debug("  Polygon has no stroke or fill. Skipping.")


    #------------------------------------------------------------------------


    def _write_polygon_filled( self, points, layer, stroke_width = 0.0 ):

        self._write_polygon_header( points, layer )

        for point in points:
            self._write_polygon_point( point )

        self._write_polygon_footer( layer, stroke_width )


    #------------------------------------------------------------------------

    def _write_polygon_outline( self, points, layer, stroke_width ):

        prior_point = None
        for point in points:

            if prior_point is not None:

                self._write_polygon_segment(
                    prior_point, point, layer, stroke_width
                )

            prior_point = point


    #------------------------------------------------------------------------

    def transform_point( self, point, flip = False ):
        ''' Transform provided point by this
        classes scale factor.
        '''

        transformed_point = svg.Point(
            ( point.x + self.translation.x ) * self.scale_factor,
            ( point.y + self.translation.y ) * self.scale_factor,
        )

        if flip:
            transformed_point.x *= -1

        if self.use_mm:
            transformed_point.x = transformed_point.x
            transformed_point.y = transformed_point.y
        else:
            transformed_point.x = int( round( transformed_point.x ) )
            transformed_point.y = int( round( transformed_point.y ) )

        return transformed_point


    #------------------------------------------------------------------------

    def write( self, cmdline="scripting" ):
        '''Write the kicad footprint file.
        The value from the command line argument
        is set in a comment in the header of the file.

        If self.file_name is not null then this will
        overwrite the target file with the data provided.
        However if it is null then all data is written
        to the string IO class (for same API as writing)
        then dumped into self.raw_file_data before the
        writer is closed.
        '''

        self._prune()

        # Must come after pruning:
        self._calculate_translation()

        if self.file_name:
            unfiltered_logger.info( "Writing module file: {}".format( self.file_name ) )
            self.output_file = open( self.file_name, 'w' )
        else:
            self.output_file = io.StringIO()

        self._write_library_intro(cmdline)

        self._write_modules()

        if self.file_name is None:
            self.raw_file_data = self.output_file.getvalue()

        self.output_file.close()
        self.output_file = None


    #------------------------------------------------------------------------

#----------------------------------------------------------------------------

class Svg2ModExportLegacy( Svg2ModExport ):
    ''' A child of Svg2ModExport that implements
    specific functionality for kicad legacy file types
    '''

    layer_map = {
        #'inkscape-name' : [ kicad-front, kicad-back ],
        'F.Cu' : [ 15, 15 ],
        'B.Cu' : [ 0, 0 ],
        'F.Adhes' : [ 17, 17 ],
        'B.Adhes' : [ 16, 16 ],
        'F.Paste' : [ 19, 19 ],
        'B.Paste' : [ 18, 18 ],
        'F.SilkS' : [ 21, 21 ],
        'B.SilkS' : [ 20, 20 ],
        'F.Mask' : [ 23, 23 ],
        'B.Mask' : [ 22, 22 ],
        'Dwgs.User' : [ 24, 24 ],
        'Cmts.User' : [ 25, 25 ],
        'Eco1.User' : [ 26, 26 ],
        'Eco2.User' : [ 27, 27 ],
        'Edge.Cuts' : [ 28, 28 ],
    }


    #------------------------------------------------------------------------

    def __init__(
        self,
        svg2mod_import,
        file_name,
        center,
        scale_factor = 1.0,
        precision = 20.0,
        use_mm = True,
        dpi = DEFAULT_DPI,
    ):
        super( Svg2ModExportLegacy, self ).__init__(
            svg2mod_import,
            file_name,
            center,
            scale_factor,
            precision,
            use_mm,
            dpi,
            pads = False,
        )

        self.include_reverse = True


    #------------------------------------------------------------------------

    def _get_layer_name( self, item_name, name, front ):

        layer_info = self.layer_map[ name ]
        layer = layer_info[ 0 ]
        if not front and layer_info[ 1 ] is not None:
            layer = layer_info[ 1 ]

        return layer


    #------------------------------------------------------------------------

    def _get_module_name( self, front = None ):

        if self.include_reverse and not front:
            return self.imported.module_name + "-rev"

        return self.imported.module_name


    #------------------------------------------------------------------------

    def _write_library_intro( self, cmdline ):

        modules_list = self._get_module_name( front = True )
        if self.include_reverse:
            modules_list += (
                "\n" +
                self._get_module_name( front = False )
            )

        units = ""
        if self.use_mm:
            units = "\nUnits mm"

        self.output_file.write( """PCBNEW-LibModule-V1  {0}{1}
$INDEX
{2}
$EndINDEX
#
# Converted using: {3}
#
""".format(
                datetime.datetime.now().strftime( "%a %d %b %Y %I:%M:%S %p %Z" ),
                units,
                modules_list,
                cmdline.replace("\\","\\\\")
            )
        )


    #------------------------------------------------------------------------

    def _write_module_header(
        self, label_size, label_pen,
        reference_y, value_y, front,
    ):

        self.output_file.write( """$MODULE {0}
Po 0 0 0 {6} 00000000 00000000 ~~
Li {0}
T0 0 {1} {2} {2} 0 {3} N I 21 "{0}"
T1 0 {5} {2} {2} 0 {3} N I 21 "{4}"
""".format(
                self._get_module_name( front ), #0
                reference_y, #1
                label_size, #2
                label_pen, #3
                self.imported.module_value, #4
                value_y, #5
                15, # Seems necessary #6
            )
        )


    #------------------------------------------------------------------------

    def _write_module_footer( self, front ):

        self.output_file.write(
            "$EndMODULE {0}\n".format( self._get_module_name( front ) )
        )


    #------------------------------------------------------------------------

    def _write_modules( self ):

        self._write_module( front = True )

        if self.include_reverse:
            self._write_module( front = False )

        self.output_file.write( "$EndLIBRARY" )


    #------------------------------------------------------------------------

    def _write_polygon_footer( self, layer, stroke_width, fill=True ):

        pass


    #------------------------------------------------------------------------

    def _write_polygon_header( self, points, layer ):

        pen = 1
        if self.use_mm:
            pen = self._convert_decimal_to_mm( pen )

        self.output_file.write( "DP 0 0 0 0 {} {} {}\n".format(
            len( points ),
            pen,
            layer
        ) )


    #------------------------------------------------------------------------

    def _write_polygon_point( self, point ):

        self.output_file.write(
            "Dl {} {}\n".format( point.x, point.y )
        )


    #------------------------------------------------------------------------

    def _write_polygon_segment( self, p, q, layer, stroke_width ):

        self.output_file.write( "DS {} {} {} {} {} {}\n".format(
            p.x, p.y,
            q.x, q.y,
            stroke_width,
            layer
        ) )


    #------------------------------------------------------------------------

    def _write_thru_hole( self, circle, layer ):

        pass

    #------------------------------------------------------------------------

#----------------------------------------------------------------------------

class Svg2ModExportLegacyUpdater( Svg2ModExportLegacy ):
    ''' A Svg2Mod exporter class that reads some settings
    from an already existing module and will append its
    changes to the file.
    '''

    #------------------------------------------------------------------------

    def __init__(
        self,
        svg2mod_import,
        file_name,
        center,
        scale_factor = 1.0,
        precision = 20.0,
        dpi = DEFAULT_DPI,
        include_reverse = True,
    ):
        self.file_name = file_name
        use_mm = self._parse_output_file()

        super( Svg2ModExportLegacyUpdater, self ).__init__(
            svg2mod_import,
            file_name,
            center,
            scale_factor,
            precision,
            use_mm,
            dpi,
        )


    #------------------------------------------------------------------------

    def _parse_output_file( self ):

        logger.info( "Parsing module file: {}".format( self.file_name ) )
        module_file = open( self.file_name, 'r' )
        lines = module_file.readlines()
        module_file.close()

        self.loaded_modules = {}
        self.post_index = []
        self.pre_index = []
        use_mm = False

        index = 0

        # Find the start of the index:
        while index < len( lines ):

            line = lines[ index ]
            index += 1
            self.pre_index.append( line )
            if line[ : 6 ] == "$INDEX":
                break

            m = re.match( r"Units[\s]+mm[\s]*", line )
            if m is not None:
                use_mm = True

        # Read the index:
        while index < len( lines ):

            line = lines[ index ]
            if line[ : 9 ] == "$EndINDEX":
                break
            index += 1
            self.loaded_modules[ line.strip() ] = []

        # Read up until the first module:
        while index < len( lines ):

            line = lines[ index ]
            if line[ : 7 ] == "$MODULE":
                break
            index += 1
            self.post_index.append( line )

        # Read modules:
        while index < len( lines ):

            line = lines[ index ]
            if line[ : 7 ] == "$MODULE":
                module_name, module_lines, index = self._read_module( lines, index )
                if module_name is not None:
                    self.loaded_modules[ module_name ] = module_lines

            elif line[ : 11 ] == "$EndLIBRARY":
                break

            else:
                raise Exception(
                    "Expected $EndLIBRARY: [{}]".format( line )
                )

        return use_mm


    #------------------------------------------------------------------------

    def _read_module( self, lines, index ):

        # Read module name:
        m = re.match( r'\$MODULE[\s]+([^\s]+)[\s]*', lines[ index ] )
        module_name = m.group( 1 )

        logger.info( "  Reading module {}".format( module_name ) )

        index += 1
        module_lines = []
        while index < len( lines ):

            line = lines[ index ]
            index += 1

            m = re.match(
                r'\$EndMODULE[\s]+' + module_name + r'[\s]*', line
            )
            if m is not None:
                return module_name, module_lines, index

            module_lines.append( line )

        raise Exception(
            "Could not find end of module '{}'".format( module_name )
        )


    #------------------------------------------------------------------------

    def _write_library_intro( self, cmdline ):

        # Write pre-index:
        self.output_file.writelines( self.pre_index )

        self.loaded_modules[ self._get_module_name( front = True ) ] = None
        if self.include_reverse:
            self.loaded_modules[
                self._get_module_name( front = False )
            ] = None

        # Write index:
        for module_name in sorted(
            self.loaded_modules.keys(),
            key = str.lower
        ):
            self.output_file.write( module_name + "\n" )

        # Write post-index:
        self.output_file.writelines( self.post_index )


    #------------------------------------------------------------------------

    def _write_preserved_modules( self, up_to = None ):

        if up_to is not None:
            up_to = up_to.lower()

        for module_name in sorted(
            self.loaded_modules.keys(),
            key = str.lower
        ):
            if up_to is not None and module_name.lower() >= up_to:
                continue

            module_lines = self.loaded_modules[ module_name ]

            if module_lines is not None:

                self.output_file.write(
                    "$MODULE {}\n".format( module_name )
                )
                self.output_file.writelines( module_lines )
                self.output_file.write(
                    "$EndMODULE {}\n".format( module_name )
                )

                self.loaded_modules[ module_name ] = None


    #------------------------------------------------------------------------

    def _write_module_footer( self, front ):

        super( Svg2ModExportLegacyUpdater, self )._write_module_footer(
            front,
        )

        # Write remaining modules:
        if not front:
            self._write_preserved_modules()


    #------------------------------------------------------------------------

    def _write_module_header(
        self,
        label_size,
        label_pen,
        reference_y,
        value_y,
        front,
    ):
        self._write_preserved_modules(
            up_to = self._get_module_name( front )
        )

        super( Svg2ModExportLegacyUpdater, self )._write_module_header(
            label_size,
            label_pen,
            reference_y,
            value_y,
            front,
        )


    #------------------------------------------------------------------------

#----------------------------------------------------------------------------

class Svg2ModExportPretty( Svg2ModExport ):
    ''' This provides functionality for the
    older kicad "pretty" footprint file formats.
    It is a child of Svg2ModExport.
    '''

    layer_map = {
        #'inkscape-name' : kicad-name,
        'F.Cu' :    "F.Cu",
        'B.Cu' :    "B.Cu",
        'F.Adhes' : "F.Adhes",
        'B.Adhes' : "B.Adhes",
        'F.Paste' : "F.Paste",
        'B.Paste' : "B.Paste",
        'F.SilkS' : "F.SilkS",
        'B.SilkS' : "B.SilkS",
        'F.Mask' :  "F.Mask",
        'B.Mask' :  "B.Mask",
        'Dwgs.User' : "Dwgs.User",
        'Cmts.User' : "Cmts.User",
        'Eco1.User' : "Eco1.User",
        'Eco2.User' : "Eco2.User",
        'Edge.Cuts' : "Edge.Cuts",
        'F.CrtYd' : "F.CrtYd",
        'B.CrtYd' : "B.CrtYd",
        'F.Fab' :   "F.Fab",
        'B.Fab' :   "B.Fab",
        'Drill.Cu': "Drill.Cu",
        'Drill.Mech': "Drill.Mech",
    }

    keepout_allowed = ['tracks','vias','pads','copperpour','footprints']


    # Breaking changes where introduced in kicad v6
    # This variable disables the drill breaking changes for v5 support
    _drill_inner_layers = False


    #------------------------------------------------------------------------

    def __init__(self, *args, **kwargs):
        super(Svg2ModExportPretty, self).__init__(*args, **kwargs)
        self._special_footer = ""
        self._extra_indent = 0

    #------------------------------------------------------------------------

    def _get_layer_name( self, item_name, name, front ):

        name = self.layer_map[ name ]

        # For pretty format layers can have attributes in the svg name
        # This validates all the layers and converts them to a json string
        # which is attached to the returned value with `:`
        attrs = {}

        # Keepout layer validation and expansion
        if name == "Keepout":
            attrs["layers"] = []
            layers = re.match(r'^(.*)\.Keepout', item_name[0]).groups()[0]
            if len(layers) == 1 and layers not in "BFI*":
                raise Exception("Unexpected keepout layer: {} in {}".format(layers, item_name[0]))
            if len(layers) == 1:
                if layers == '*':
                    attrs['layers'] = ['F','B','I']
                else:
                    attrs['layers'] = [layers]
            else:
                for layer in layers:
                    if layer not in "FBI":
                        raise Exception("Unexpected keepout layer: {} in {}".format(layer, item_name[0]))
                    if layer != '&':
                        attrs['layers'].append(layer)

        # All attributes with the exception of keepout layers is validated
        if len(item_name) == 2 and item_name[1]:
            for arg in item_name[1].split(';'):
                arg = arg.strip(' ,:')
                # This is used in Svg2ModExportLatest as it is a breaking change
                if name == "Keepout" and re.match(r'^allowed:\w+', arg, re.I):
                    attrs["allowed"] = []
                    for allowed in arg.lower().split(":", 1)[1].split(','):
                        if allowed in self.keepout_allowed:
                            attrs["allowed"].append(allowed)
                        else:
                            logger.warning("Invalid allowed option in keepout: {} in {}".format(allowed, arg))
                elif re.match(r'^\w+\.Cu', name) and re.match(r'^pad(:(\d+|mask|paste))?', arg, re.I):
                    if arg.lower() == "pad":
                        attrs["copper_pad"] = True
                    else:
                        ops = arg.split(":", 1)[1]
                        for opt in ops.split(','):
                            if re.match(r'^\d+$', opt):
                                attrs["copper_pad"] = int(opt)
                            elif opt.lower() == "mask" and name != "Drill.Cu":
                                attrs["pad_mask"] = True
                                if not attrs.get("copper_pad"):
                                    attrs["copper_pad"] = True
                            elif opt.lower() == "paste" and name != "Drill.Cu":
                                attrs["pad_paste"] = True
                                if not attrs.get("copper_pad"):
                                    attrs["copper_pad"] = True
                            else:
                                logger.warning("Invalid pad option '{}' for layer {}".format(opt, name))
                else:
                    logger.warning("Unexpected option: {} for {}".format(arg, item_name[0]))
        if attrs:
            return name+":"+json.dumps(attrs)
        return name


    #------------------------------------------------------------------------

    def _get_module_name( self, front = None ):

        return self.imported.module_name


    #------------------------------------------------------------------------

    def _write_library_intro( self, cmdline ):

        self.output_file.write( """(module {0} (layer F.Cu) (tedit {1:8X})
  (attr virtual)
  (descr "{2}")
  (tags {3})
""".format(
                self.imported.module_name, #0
                int( round( #1
                    os.path.getctime( self.imported.file_name ) if self.imported.file_name else time.time()
                ) ),
                "Converted using: {}".format( cmdline.replace("\\", "\\\\") ), #2
                "svg2mod", #3
            )
        )


    #------------------------------------------------------------------------

    def _write_module_footer( self, front ):

        self.output_file.write( "\n)" )


    #------------------------------------------------------------------------

    def _write_module_header(
        self, label_size, label_pen,
        reference_y, value_y, front,
    ):
        if front:
            side = "F"
        else:
            side = "B"

        self.output_file.write(
"""  (fp_text reference {0} (at 0 {1}) (layer {2}.SilkS) hide
    (effects (font (size {3} {3}) (thickness {4})))
  )
  (fp_text value {5} (at 0 {6}) (layer {2}.SilkS) hide
    (effects (font (size {3} {3}) (thickness {4})))
  )""".format(

                self._get_module_name(), #0
                reference_y, #1
                side, #2
                label_size, #3
                label_pen, #4
                self.imported.module_value, #5
                value_y, #6
            )
        )


    #------------------------------------------------------------------------

    def _write_modules( self ):

        self._write_module( front = True )


    #------------------------------------------------------------------------

    def _write_polygon_filled( self, points, layer, stroke_width = 0):
        self._write_polygon_header( points, layer, stroke_width)

        for point in points:
            self._write_polygon_point( point )

        self._write_polygon_footer( layer, stroke_width )

    #------------------------------------------------------------------------

    def _write_polygon_footer( self, layer, stroke_width, fill=True ):

        #Format option #2 is expected, but only used in Svg2ModExportLatest
        if self._special_footer:
            self.output_file.write(self._special_footer.format(
                layer.split(":", 1)[0], stroke_width, "" #2
            ))
        else:
            self.output_file.write(
                "    )\n    (layer {})\n    (width {}){}\n  )".format(
                    layer.split(":", 1)[0], stroke_width, "" #3
                )
            )
        self._special_footer = ""
        self._extra_indent = 0


    #------------------------------------------------------------------------

    def _write_polygon_header( self, points, layer, stroke_width):

        l_name = layer
        options = {}
        try:
            l_name, options = layer.split(":", 1)
            options = json.loads(options)
        except ValueError:pass

        create_pad = (self.convert_pads and l_name.find("Cu") == 2) or options.get("copper_pad")

        if stroke_width == 0:
            stroke_width = MINIMUM_SIZE

        if l_name == "Keepout":
            self._extra_indent = 1
            layers = ["*"]
            if len(options["layers"]) == 3:
                layers = ["*"]
            elif "I" not in options["layers"]:
                layers = ["&".join(options["layers"])]
            else:
                options["layers"].remove("I")
                layers = options["layers"][:] + ['In{}'.format(i) for i in range(1,31)]


            self.output_file.write( '''\n  (zone (net 0) (net_name "") (layers "{0}.Cu") (hatch edge {1:.6f})
    (connect_pads (clearance 0))
    (min_thickness {2:.6f})
    (keepout ({3}allowed))
    (fill (thermal_gap {1:.6f}) (thermal_bridge_width {1:.6f}))
    (polygon
      (pts\n'''.format(
                    '.Cu" "'.join(layers), #0
                    stroke_width, #1
                    stroke_width/2, #2
                    "allowed) (".join(
                        [i+" "+(
                            "not_" if not options.get("allowed") or i not in options["allowed"] else ""
                        ) for i in self.keepout_allowed]
                    ), #3
                )
            )
            self._special_footer = "      )\n    )\n  )"
        elif create_pad:
            pad_number = "" if not options.get("copper_pad") or isinstance(options["copper_pad"], bool) else str(options.get("copper_pad"))
            layer = l_name
            if options.get("pad_mask"):
                layer += " {}.Mask".format(l_name.split(".", 1)[0])
            if options.get("pad_paste"):
                layer += " {}.Paste".format(l_name.split(".", 1)[0])
            self._extra_indent = 1

            self._special_footer = "\n  )"

            self.output_file.write( '''\n  (pad "{0}" smd custom (at {1} {2}) (size {3:.6f} {3:.6f}) (layers {4})
    (zone_connect 0)
    (options (clearance outline) (anchor circle))'''.format(
                    pad_number, #0
                    points[0].x, #1
                    points[0].y, #2
                    stroke_width, #3
                    layer, #4
                )
            )
            # Pads primitives with 2 or less points crash kicad
            if len(points) >= 2:
                self.output_file.write('''\n    (primitives\n      (gr_poly (pts \n''')
                self._special_footer = "      )\n    (width {}){{2}})\n  ))".format(stroke_width)

                originx = points[0].x
                originy = points[0].y
                for point in points:
                    point.x = point.x-originx
                    point.y = point.y-originy
            else:
                for point in points[:]:
                    points.remove(point)
        else:
            self.output_file.write( "\n  (fp_poly\n    (pts \n" )


    #------------------------------------------------------------------------

    def _write_polygon_point( self, point ):

        if self._extra_indent:
            self.output_file.write("  "*self._extra_indent)

        self.output_file.write(
            "      (xy {} {})\n".format( point.x, point.y )
        )


    #------------------------------------------------------------------------

    def _write_polygon_segment( self, p, q, layer, stroke_width ):
        l_name = layer
        options = {}
        try:
            l_name, options = layer.split(":", 1)
            options = json.loads(options)
        except ValueError:pass

        create_pad = (self.convert_pads and l_name.find("Cu") == 2) or options.get("copper_pad")

        if stroke_width == 0:
            stroke_width = MINIMUM_SIZE

        if create_pad:
            pad_number = "" if not options.get("copper_pad") or isinstance(options["copper_pad"], bool) else str(options.get("copper_pad"))
            layer = l_name
            if options.get("pad_mask"):
                layer += " {}.Mask".format(l_name.split(".", 1)[0])
            if options.get("pad_paste"):
                layer += " {}.Paste".format(l_name.split(".", 1)[0])

            # There are major performance issues when multiple line primitives are in the same pad
            self.output_file.write( '''\n  (pad "{0}" smd custom (at {1} {2}) (size {3:.6f} {3:.6f}) (layers {4})
    (zone_connect 0)
    (options (clearance outline) (anchor circle))
    (primitives\n      (gr_line (start 0 0) (end {5} {6}) (width {3}))
  ))'''.format(
                    pad_number, #0
                    p.x, #1
                    p.y, #2
                    stroke_width, #3
                    layer, #4
                    q.x - p.x, #5
                    q.y - p.y, #6
                )
            )
        else:

            self.output_file.write(
                """\n  (fp_line
        (start {} {}) (end {} {})
        (layer {}) (width {})
    )""".format(
                    p.x, p.y,
                    q.x, q.y,
                    layer.split(':',1)[0],
                    stroke_width,
                )
            )

    #------------------------------------------------------------------------

    def _write_thru_hole( self, circle, layer ):

        l_name = layer
        options = {}
        try:
            l_name, options = layer.split(":", 1)
            options = json.loads(options)
        except ValueError:pass

        plated = l_name == "Drill.Cu"
        pad_number = ""
        if plated and options.get("copper_pad") and not isinstance(options["copper_pad"], bool):
            pad_number = str(options.get("copper_pad"))

        rad = circle.rx * self.scale_factor
        drill = rad * 2

        size = circle.style.get("stroke-width") * self.scale_factor
        if size and plated:
            drill -= size
            size = size + ( rad * 2 )
        elif size:
            size = size + ( rad * 2 )
            drill = size
        else:
            size = rad

        center = self.transform_point(circle.center)

        self.output_file.write(
            '\n  (pad "{0}" {1}thru_hole circle (at {2} {3}) (size {4} {4}) (drill {5}) (layers *.Mask{6}) {7})'.format(
                pad_number, #0
                "" if plated else "np_", #1
                center.x, #2
                center.y, #3
                size, #4
                drill, #5
                " *.Cu" if plated else "", #6
                "(remove_unused_layers) (keep_end_layers)" if plated and self._drill_inner_layers else "", #7
            )
        )

    #------------------------------------------------------------------------

#----------------------------------------------------------------------------


class Svg2ModExportLatest(Svg2ModExportPretty):
    ''' This provides functionality for the newer kicad
    "pretty" footprint file formats introduced in kicad v6.
    It is a child of Svg2ModExport.
    '''

    layer_map = {
        #'inkscape-name' : kicad-name,
        'F.Cu' :    "F.Cu",
        'B.Cu' :    "B.Cu",
        'F.Adhes' : "F.Adhes",
        'B.Adhes' : "B.Adhes",
        'F.Paste' : "F.Paste",
        'B.Paste' : "B.Paste",
        'F.SilkS' : "F.SilkS",
        'B.SilkS' : "B.SilkS",
        'F.Mask' :  "F.Mask",
        'B.Mask' :  "B.Mask",
        'Dwgs.User' : "Dwgs.User",
        'Cmts.User' : "Cmts.User",
        'Eco1.User' : "Eco1.User",
        'Eco2.User' : "Eco2.User",
        'Edge.Cuts' : "Edge.Cuts",
        'F.CrtYd' : "F.CrtYd",
        'B.CrtYd' : "B.CrtYd",
        'F.Fab' :   "F.Fab",
        'B.Fab' :   "B.Fab",
        'Drill.Cu': "Drill.Cu",
        'Drill.Mech': "Drill.Mech",
        r'\S+\.Keepout': "Keepout"
    }

    # Breaking changes where introduced in kicad v6
    # This variable enables the drill breaking changes for v5 support
    _drill_inner_layers = True

    #------------------------------------------------------------------------

    def _write_polygon_outline( self, points, layer, stroke_width = 0):
        self._write_polygon_header( points, layer, stroke_width)

        for point in points:
            self._write_polygon_point( point )

        self._write_polygon_footer( layer, stroke_width, fill=False )

    #------------------------------------------------------------------------

    def _write_polygon_footer( self, layer, stroke_width, fill=True ):

        if self._special_footer:
            self.output_file.write(self._special_footer.format(
                layer.split(":", 1)[0], stroke_width,
                " (fill none)" if not fill else ""
            ))
        else:
            self.output_file.write(
                "    )\n    (layer {})\n    (width {}){}\n  )".format(
                    layer.split(":", 1)[0], stroke_width,
                    " (fill none)" if not fill else ""
                )
            )
        self._special_footer = ""
        self._extra_indent = 0



    #------------------------------------------------------------------------

#----------------------------------------------------------------------------



# Copyright (C) 2021 -- svg2mod developers < GitHub.com / svg2mod >

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

'''
Svg2ModImport is responsible for basic parsing
of svg layers to be used with an instance of
Svg2ModExport.
'''


from svg2mod import svg
from svg2mod.coloredlogger import logger, unfiltered_logger

#----------------------------------------------------------------------------

class Svg2ModImport:
    ''' An importer class to read in target svg,
    parse it, and keep only layers on interest.
    '''

    #------------------------------------------------------------------------

    def _prune_hidden( self, items = None ):

        if items is None:

            items = self.svg.items

        for item in items[:]:

            if hasattr(item, "hidden") and item.hidden:
                if hasattr(item, "name") and item.name:
                    logger.warning("Ignoring hidden SVG item: {}".format( item.name ) )
                items.remove(item)

            if hasattr(item, "items") and item.items:
                self._prune_hidden( item.items )

    #------------------------------------------------------------------------

    def __init__( self, file_name=None, module_name="svg2mod", module_value="G***", ignore_hidden=False, force_layer=None):

        self.file_name = file_name
        self.module_name = module_name
        self.module_value = module_value
        self.ignore_hidden = ignore_hidden

        if file_name:
            unfiltered_logger.info( "Parsing SVG..." )

            self.svg = svg.parse( file_name )
            logger.info("Document scaling: {} units per pixel".format(self.svg.viewport_scale))
        if force_layer:
            new_layer = svg.Group()
            new_layer.name = force_layer
            new_layer.items = self.svg.items[:]
            self.svg.items = [new_layer]
        if self.ignore_hidden:
            self._prune_hidden()


    #------------------------------------------------------------------------

#----------------------------------------------------------------------------



# Copyright (C) 2022 -- svg2mod developers < GitHub.com / svg2mod >

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

'''
Helper classes to combine and calculate the points
from a svg object into a single continuous line
'''

import copy
from typing import List, Tuple

from svg2mod import svg
from svg2mod.coloredlogger import logger

#----------------------------------------------------------------------------

class LineSegment:
    '''Kicad can only draw straight lines.
    It is designed to have extra functions to help
    calculate intersections.
    '''

    #------------------------------------------------------------------------

    @staticmethod
    def _on_segment( p, q, r ):
        """ Given three collinear points p, q, and r, check if
            point q lies on line segment pr. """

        if (
            q.x <= max( p.x, r.x ) and
            q.x >= min( p.x, r.x ) and
            q.y <= max( p.y, r.y ) and
            q.y >= min( p.y, r.y )
        ):
            return True

        return False


    #------------------------------------------------------------------------

    @staticmethod
    def _orientation( p, q, r ):
        """ Find orientation of ordered triplet (p, q, r).
            Returns following values
            0 --> p, q and r are collinear
            1 --> Clockwise
            2 --> Counterclockwise
        """

        val = (
            ( q.y - p.y ) * ( r.x - q.x ) -
            ( q.x - p.x ) * ( r.y - q.y )
        )

        if val == 0: return 0
        if val > 0: return 1
        return 2

    #------------------------------------------------------------------------

    @staticmethod
    def vertical_intersection(p: svg.Point, q: svg.Point, r: float) -> svg.Point:
        '''This is used for the in-lining algorithm
        it finds a point on a line p -> q where x = r
        '''
        if p.x == q.x:
            return min([p,q], key=lambda v: v.y)
        if r == p.x: return p
        if r == q.x: return q
        return svg.Point(r, (p.y-q.y)*(r-q.x)/(p.x-q.x)+q.y)


    #------------------------------------------------------------------------

    def __init__( self, p = None, q = None ):

        self.p = p
        self.q = q


    #------------------------------------------------------------------------

    def connects( self, segment: 'LineSegment' ) -> bool:
        ''' Return true if provided segment shares
        endpoints with the current segment
        '''

        if self.q == segment.p: return True
        if self.q == segment.q: return True
        if self.p == segment.p: return True
        if self.p == segment.q: return True
        return False

    #------------------------------------------------------------------------

    def on_line(self, point: svg.Point) -> bool:
        '''Returns true if the point is on the line.
            Adapted from:
            https://stackoverflow.com/questions/36487156/javascript-determine-if-a-point-resides-above-or-below-a-line-defined-by-two-poi
        '''
        return not (self.p.x-self.q.x)*(point.y-self.q.y) - (self.p.y-self.q.y)*(point.x-self.q.x)

    #------------------------------------------------------------------------

    def intersects( self, segment: 'LineSegment' ) -> bool:
        """ Return true if line segments 'p1q1' and 'p2q2' intersect.
            Adapted from:
              http://www.geeksforgeeks.org/check-if-two-given-line-segments-intersect/
        """

        # Find the four orientations needed for general and special cases:
        o1 = self._orientation( self.p, self.q, segment.p )
        o2 = self._orientation( self.p, self.q, segment.q )
        o3 = self._orientation( segment.p, segment.q, self.p )
        o4 = self._orientation( segment.p, segment.q, self.q )

        return (

            # General case:
            ( o1 != o2 and o3 != o4 )

            or

            # p1, q1 and p2 are collinear and p2 lies on segment p1q1:
            ( o1 == 0 and self._on_segment( self.p, segment.p, self.q ) )

            or

            # p1, q1 and p2 are collinear and q2 lies on segment p1q1:
            ( o2 == 0 and self._on_segment( self.p, segment.q, self.q ) )

            or

            # p2, q2 and p1 are collinear and p1 lies on segment p2q2:
            ( o3 == 0 and self._on_segment( segment.p, self.p, segment.q ) )

            or

            # p2, q2 and q1 are collinear and q1 lies on segment p2q2:
            ( o4 == 0 and self._on_segment( segment.p, self.q, segment.q ) )
        )


    #------------------------------------------------------------------------

    def q_next( self, q:svg.Point ):
        '''Shift segment endpoints so self.q is self.p
        and q is the new self.q
        '''

        self.p = self.q
        self.q = q

    #------------------------------------------------------------------------

    def __eq__(self, other):
        return (
            isinstance(other, LineSegment) and
            other.p.x == self.p.x and other.p.y == self.p.y and
            other.q.x == self.q.x and other.q.y == self.q.y
        )

    #------------------------------------------------------------------------

#----------------------------------------------------------------------------

class PolygonSegment:
    ''' A polygon should be a collection of segments
    creating an enclosed or manifold shape.
    This class provides functionality to find overlap
    points between a segment and it's self as well as
    identify if another polygon rests inside of the
    closed area of it's self.

    When initializing this class it will remove duplicate points in a row.
    '''

    #------------------------------------------------------------------------

    def __init__( self, points:List):

        self.points = [points[0]]

        for point in points:
            if self.points[-1] != point:
                self.points.append(point)


        self.bbox = None
        self.calc_bbox()


    #------------------------------------------------------------------------

    def _set_points(self, points: List[svg.Point]):
        self.points = points[:]

    #------------------------------------------------------------------------

    def _find_insertion_point( self, hole: 'PolygonSegment', holes: list, other_insertions: list ):
        ''' KiCad will not "pick up the pen" when moving between a polygon outline
        and holes within it, so we search for a pair of points connecting the
        outline (self) or other previously inserted points to the hole such
        that the connecting segment will not cross the visible inner space
        within any hole.
        '''

        highest_point = max(hole.points, key=lambda v: v.y)
        vertical_line = LineSegment(highest_point, svg.Point(highest_point.x, self.bbox[1].y+1))

        intersections = {self: self.intersects(vertical_line, False, count_intersections=True, get_points=True)}
        for _,h,__ in other_insertions:
            if h.bbox[0].x < highest_point.x and h.bbox[1].x > highest_point.x:
                intersections[h] = h.intersects(vertical_line, False, count_intersections=True, get_points=True)

        best = [self, intersections[self][0]]
        best.append(LineSegment.vertical_intersection(best[1][0], best[1][1], highest_point.x))
        for path in intersections:
            for p,q in intersections[path]:
                pnt = LineSegment.vertical_intersection(p, q, highest_point.x)
                if pnt.y < best[2].y:
                    best = [path, (p,q), pnt]

        if best[2] != best[1][0] and best[2] != best[1][1]:
            p = best[0].points.index(best[1][0])
            q = best[0].points.index(best[1][1])
            ip = p if p < q else q
            best[0]._set_points(best[0].points[:ip+1] + [best[2]] + best[0].points[ip+1:])

        return (best[2], hole, highest_point)

    #------------------------------------------------------------------------

    def points_starting_on_index( self, index: int ) -> List[svg.Point]:
        ''' Return the list of ordered points starting on the given index, ensuring
        that the first and last points are the same.
        '''

        points = self.points

        if index > 0:

            # Strip off end point, which is a duplicate of the start point:
            points = points[ : -1 ]

            points = points[ index : ] + points[ : index ]

            points.append(
                svg.Point( points[ 0 ].x, points[ 0 ].y )
            )

        return points


    #------------------------------------------------------------------------

    def inline( self, segments: List[svg.Point] ) -> List[svg.Point]:
        ''' Return a list of points with the given polygon segments (paths) inlined. '''

        if len( segments ) < 1:
            return self.points

        logger.debug( "  Inlining {} segments...".format( len( segments ) ) )

        segments.sort(reverse=True, key=lambda h: h.bbox[1].y)

        all_segments = segments[ : ] + [ self ]
        insertions = []

        # Find the insertion point for each hole:
        for hole in segments:

            insertion = self._find_insertion_point(
                hole, all_segments, insertions
            )
            if insertion is not None:
                insertions.append( insertion )

        # Prevent returned points from affecting original object
        points = copy.deepcopy(self.points)

        for insertion in insertions:

            ip = points.index(insertion[0])
            hole = insertion[1].points_starting_on_index(insertion[1].points.index(insertion[2]))

            if (
                points[ ip ].x == hole[ 0 ].x and
                points[ ip ].y == hole[ 0 ].y
            ):
                # The point at the insertion point is duplicated so any action on that will affect both
                points = points[:ip] + [copy.copy(points[ip])] + hole[ 1 : -1 ] + points[ip:]
            else:
                # The point at the insertion point is duplicated so any action on that will affect both
                points = points[:ip] + [copy.copy(points[ip])] + hole + points[ip:]

        return points


    #------------------------------------------------------------------------

    def intersects( self, line_segment: LineSegment, check_connects:bool , count_intersections=False, get_points=False):
        '''Check to see if line_segment intersects with any
        segments of the polygon. Default return True/False

        If check_connects is True then it will skip intersections
        that share endpoints with line_segment.

        count_intersections will return the number of intersections
        with the polygon.

        get_points returns a tuple of the line that intersects
        with line_segment. count_intersection in combination will
        return a list of tuples of line segments.
        '''

        hole_segment = LineSegment()

        intersections = 0
        intersect_segs = []
        virt_line = LineSegment()

        # Check each segment of other hole for intersection:
        for point in self.points:

            hole_segment.q_next( point )

            if hole_segment.p is not None:

                if ( check_connects and line_segment.connects( hole_segment )):
                    continue

                if line_segment.intersects( hole_segment ):

                    if count_intersections:
                        if get_points:
                            intersect_segs.append((hole_segment.p, hole_segment.q))
                        else:
                            # If a point is on the line segment we need to see if the
                            # simplified "virtual" line crosses the line segment.

                            # Set the endpoints if they are of the line segment
                            if line_segment.on_line(hole_segment.q):
                                if not line_segment.on_line(hole_segment.p):
                                    virt_line.p = hole_segment.p
                            elif line_segment.on_line(hole_segment.p):
                                virt_line.q = hole_segment.q

                            # No points are on the line segment
                            else:
                                intersections += 1
                                virt_line = LineSegment()

                            # The virtual line is complete check for intersections
                            if virt_line.p and virt_line.q:
                                if virt_line.intersects(line_segment):
                                    intersections += 1
                                virt_line = LineSegment()

                    elif get_points:
                        return hole_segment.p, hole_segment.q
                    else:
                        return True

        if count_intersections:
            return intersect_segs if get_points else intersections
        if get_points and not check_connects:
            return ()
        return False


    #------------------------------------------------------------------------

    def process( self, transformer, flip, fill ):
        ''' Apply all transformations, then remove duplicate
        consecutive points along the path.
        '''

        points = []
        for point in self.points:

            point = transformer.transform_point( point, flip )

            if (
                len( points ) < 1 or
                point.x != points[ -1 ].x or
                point.y != points[ -1 ].y
            ):
                points.append( point )

        if (
            points[ 0 ].x != points[ -1 ].x or
            points[ 0 ].y != points[ -1 ].y
        ):

            if fill:
                points.append( svg.Point(
                    points[ 0 ].x,
                    points[ 0 ].y,
                ) )

        self.points = points
        self.calc_bbox()


    #------------------------------------------------------------------------

    def calc_bbox(self) -> Tuple[svg.Point, svg.Point]:
        '''Calculate bounding box of self'''
        self.bbox =  (
            svg.Point(min(self.points, key=lambda v: v.x).x, min(self.points, key=lambda v: v.y).y),
            svg.Point(max(self.points, key=lambda v: v.x).x, max(self.points, key=lambda v: v.y).y),
        )

    #------------------------------------------------------------------------

    def are_distinct(self, polygon):
        ''' Checks if the supplied polygon either contains or insets our bounding box'''
        distinct = True

        smaller = min([self, polygon], key=lambda p: svg.Segment(p.bbox[0], p.bbox[1]).length())
        larger = self if smaller == polygon else polygon

        if (
            larger.bbox[0].x < smaller.bbox[0].x and
            larger.bbox[0].y < smaller.bbox[0].y and
            larger.bbox[1].x > smaller.bbox[1].x and
            larger.bbox[1].y > smaller.bbox[1].y
        ):
            distinct = False

        # Check number of horizontal intersections. If the number is odd then it the smaller polygon
        # is contained. If the number is even then the polygon is outside of the larger polygon
        if not distinct:
            tline = LineSegment(smaller.points[0], svg.Point(larger.bbox[1].x+1, smaller.points[0].y))
            distinct = bool((larger.intersects(tline, False, True) + 1)%2)

        return distinct
    #------------------------------------------------------------------------


#----------------------------------------------------------------------------




# Copyright (C) 2013 -- CJlano < cjlano @ free.fr >
# Copyright (C) 2022 -- svg2mod developers < GitHub.com / svg2mod >

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
'''
A SVG parser with tools to convert an XML svg file
to objects that can be simplified into points.
'''

import copy
import inspect
import itertools
import json
import logging
import math
import operator
import os
import platform
import re
import sys
import xml.etree.ElementTree as etree
from typing import List, Tuple

from fontTools.misc import loggingTools
from fontTools.pens.svgPathPen import SVGPathPen
from fontTools.ttLib import ttFont
from svg2mod.coloredlogger import logger

from .geometry import Angle, Bezier, MoveTo, Point, Segment, simplify_segment

svg_ns = '{http://www.w3.org/2000/svg}'

# Regex commonly used
number_re = r'[-+]?(?:\d+(?:\.\d*)?|\.\d+)(?:[eE][-+]?\d+)?'
unit_re = r'em|ex|px|in|cm|mm|pt|pc|%'

# Unit converter
unit_convert = {
        None: 1,           # Default unit (same as pixel)
        'px': 1,           # px: pixel. Default SVG unit
        'em': 10,          # 1 em = 10 px FIXME
        'ex': 5,           # 1 ex =  5 px FIXME
        'in': 96,          # 1 in = 96 px
        'cm': 96 / 2.54,   # 1 cm = 1/2.54 in
        'mm': 96 / 25.4,   # 1 mm = 1/25.4 in
        'pt': 96 / 72.0,   # 1 pt = 1/72 in
        'pc': 96 / 6.0,    # 1 pc = 1/6 in
        '%' :  1 / 100.0   # 1 percent
        }

# Logging spammed 'Unable to find font because no font was specified.'
# this allows it to only print the error once before muting it for that run.
_font_warning_sent = False


class Transformable:
    '''Abstract class for objects that can be geometrically drawn & transformed'''

    # This list is all styles that should have the transformation matrix applied
    transformable_styles = ["stroke-width"]

    def __init__(self, elt=None, parent_styles=None):
        # a 'Transformable' is represented as a list of Transformable items
        self.items = []
        self.id = hex(id(self))
        # Unit transformation matrix on init
        self.matrix = Matrix()
        self.scalex = 1
        self.scaley = 1
        self.style = {} if not parent_styles and not isinstance(parent_styles, dict) else parent_styles.copy()
        self.rotation = 0
        self.viewport = Point(800, 600) # default viewport is 800x600
        if elt is not None:
            self.id = elt.get('id', self.id)

            # parse styles and save as dictionary.
            if elt.get('style'):
                for style in elt.get('style').split(";"):
                    if style.find(":") == -1:
                        self.style[name] = None
                    else:
                        nv = style.split(":")
                        name = nv[ 0 ].strip()
                        value = nv[ 1 ].strip()
                        if name in self.transformable_styles:
                            value = list(re.search(r'(\d+\.?\d*)(\D+)?', value).groups())
                            self.style[name] = float(value[0])
                            if value[1] and value[1] not in unit_convert:
                                logger.warning("Style '{}' has an unexpected unit: {}".format(style, value[1]))
                        else:
                            self.style[name] = value

            # Parse transform attribute to update self.matrix
            self.get_transformations(elt)

        if self.style.get("display") == "none":
            self.hidden = True

    def bbox(self):
        '''Bounding box of all points'''
        bboxes = [x.bbox() for x in self.items]
        if len( bboxes ) < 1:
            return (Point(0, 0), Point(0, 0))
        xmin = min([b[0].x for b in bboxes])
        xmax = max([b[1].x for b in bboxes])
        ymin = min([b[0].y for b in bboxes])
        ymax = max([b[1].y for b in bboxes])

        return (Point(xmin,ymin), Point(xmax,ymax))

    # Parse transform field
    def get_transformations(self, elt):
        '''Take an xml element and parse transformation commands
        then apply the matrix and set any needed variables
        '''
        t = elt.get('transform')
        if t is None: return

        svg_transforms = [
                'matrix', 'translate', 'scale', 'rotate', 'skewX', 'skewY']

        # match any SVG transformation with its parameter (until final parenthesis)
        # [^)]*    == anything but a closing parenthesis
        # '|'.join == OR-list of SVG transformations
        transforms = re.findall(
                '|'.join([x + r'[^)]*\)' for x in svg_transforms]), t)

        for t in transforms:
            op, arg = t.split('(')
            op = op.strip()
            # Keep only numbers
            arg = [float(x) for x in re.findall(number_re, arg)]
            logger.debug('transform: ' + op + ' '+ str(arg))

            if op == 'matrix':
                self.matrix *= Matrix(arg)

            if op == 'translate':
                tx = arg[0]
                if len(arg) == 1: ty = 0
                else: ty = arg[1]
                self.matrix *= Matrix([1, 0, 0, 1, tx, ty])

            if op == 'scale':
                sx = arg[0]
                if len(arg) == 1: sy = sx
                else: sy = arg[1]
                self.scalex *= sx
                self.scaley *= sy
                self.matrix *= Matrix([sx, 0, 0, sy, 0, 0])

            if op == 'rotate':
                self.rotation += arg[0]
                cosa = math.cos(math.radians(arg[0]))
                sina = math.sin(math.radians(arg[0]))
                if len(arg) != 1:
                    tx, ty = arg[1:3]
                    self.matrix *= Matrix([1, 0, 0, 1, tx, ty])
                self.matrix *= Matrix([cosa, sina, -sina, cosa, 0, 0])
                if len(arg) != 1:
                    self.matrix *= Matrix([1, 0, 0, 1, -tx, -ty])

            if op == 'skewX':
                tana = math.tan(math.radians(arg[0]))
                self.matrix *= Matrix([1, 0, tana, 1, 0, 0])

            if op == 'skewY':
                tana = math.tan(math.radians(arg[0]))
                self.matrix *= Matrix([1, tana, 0, 1, 0, 0])

    def transform_styles(self, matrix):
        '''Any style in this classes transformable_styles
        will be scaled by the provided matrix.
        '''
        for style in self.transformable_styles:
            if self.style.get(style):
                self.style[style] = float(self.style[style]) * ((matrix.xscale()+matrix.yscale())/2)


    def transform(self, matrix=None):
        '''Apply the provided matrix. Default (None)
        If no matrix is supplied then recursively apply
        it's already existing matrix to all items.
        '''
        if matrix is None:
            matrix = self.matrix
        else:
            matrix *= self.matrix
        self.transform_styles(matrix)
        for x in self.items:
            x.transform(matrix)

    def length(self, v, mode='xy'):
        '''Return generic 2 dimensional length of svg element'''
        # Handle empty (non-existing) length element
        if v is None:
            return 0

        # Get length value
        m = re.search(number_re, v)
        if m: value = m.group(0)
        else: raise TypeError(v + 'is not a valid length')

        # Get length unit
        m = re.search(unit_re, v)
        if m: unit = m.group(0)
        else: unit = None

        if unit == '%':
            if mode == 'x':
                return float(value) * unit_convert[unit] * self.viewport.x
            if mode == 'y':
                return float(value) * unit_convert[unit] * self.viewport.y
            if mode == 'xy':
                return float(value) * unit_convert[unit] * self.viewport.x # FIXME

        return float(value) * unit_convert[unit]

    def xlength(self, x):
        '''Length of element's x component'''
        return self.length(x, 'x')
    def ylength(self, y):
        '''Length of element's y component'''
        return self.length(y, 'y')

    def flatten(self):
        '''Flatten the SVG objects nested list into a flat (1-D) list,
        removing Groups'''
        # http://rightfootin.blogspot.fr/2006/09/more-on-python-flatten.html
        # Assigning a slice a[i:i+1] with a list actually replaces the a[i]
        # element with the content of the assigned list
        i = 0
        flat = copy.deepcopy(self.items)
        while i < len(flat):
            while isinstance(flat[i], Group):
                flat[i:i+1] = flat[i].items
            i += 1
        return flat

class Svg(Transformable):
    '''SVG class: use parse to parse a file'''
    # class Svg handles the <svg> tag
    # tag = 'svg'

    def __init__(self, filename=None):
        self.viewport_scale = 1
        Transformable.__init__(self)
        if filename:
            self.parse(filename)

    def parse(self, filename:str):
        '''Read provided svg xml file and
        append all svg element to items list
        '''
        self.filename = filename
        tree = etree.parse(filename)
        self.root = tree.getroot()
        if self.root.tag != svg_ns + 'svg':
            raise TypeError('file %s does not seem to be a valid SVG file', filename)

        # Create a top Group to group all other items (useful for viewBox elt)
        top_group = Group()
        self.items.append(top_group)

        # SVG dimension
        width = self.xlength(self.root.get('width'))
        height = self.ylength(self.root.get('height'))

        # update viewport
        top_group.viewport = Point(width, height)

        # viewBox
        if self.root.get('viewBox') is not None:
            view_box = re.findall(number_re, self.root.get('viewBox'))

            # If the document somehow doesn't have dimensions get if from viewBox
            if self.root.get('width') is None or self.root.get('height') is None:
                width = float(view_box[2])
                height = float(view_box[3])
                logger.warning("Unable to find width or height properties. Using viewBox.")

            sx = width / float(view_box[2])
            sy = height / float(view_box[3])
            tx = -float(view_box[0])
            ty = -float(view_box[1])
            self.viewport_scale = round(float(view_box[2])/width, 6)
            top_group.matrix = Matrix([sx, 0, 0, sy, tx, ty])
        if ( self.root.get("width") is None or self.root.get("height") is None ) \
                and self.root.get("viewBox") is None:
            logger.critical("Fatal Error: Unable to find SVG dimensions. Exiting.")
            sys.exit(-1)

        # Parse XML elements hierarchically with groups <g>
        top_group.append(self.root)

        self.transform()

    def title(self):
        '''Returns svg title if exists. Otherwise try to return filename'''
        t = self.root.find(svg_ns + 'title')
        if t is not None:
            return t
        return os.path.splitext(os.path.basename(self.filename))[0]

    def json(self):
        '''Return a dictionary of children items'''
        return self.items


class Group(Transformable):
    '''Handle svg <g> elements
    The name and hidden attributes are stored in self.name
    and self.hidden respectively. These can be manually set
    if object is not initialized with an xml element.
    '''
    # class Group handles the <g> tag
    tag = 'g'

    def __init__(self, elt=None, *args, **kwargs):
        Transformable.__init__(self, elt, *args, **kwargs)

        self.name = ""
        if elt is not None:
            for ident, value in elt.attrib.items():

                ident = self.parse_name( ident )
                if ident[ "name" ] == "label":
                    self.name = value

    @staticmethod
    def parse_name( tag ):
        '''Read and return name from xml data'''
        m = re.match( r'({(.+)})?(.+)', tag )
        return {
            'namespace' : m.group( 2 ),
            'name' : m.group( 3 ),
        }

    def append(self, element):
        '''Convert and append xml element(s) to items list
        element is expected to be iterable.
        If an svg non xml object needs to be appended
        then interface directly with the items list:
                group.items.append(svg_object)
        '''
        for elt in element:
            elt_class = svgClass.get(elt.tag, None)
            if elt_class is None:
                logger.debug('No handler for element %s' % elt.tag)
                continue
            # instantiate elt associated class (e.g. <path>: item = Path(elt)
            item = elt_class(elt, parent_styles=self.style)
            # Apply group matrix to the newly created object
            # Actually, this is effectively done in Svg.__init__() through call to
            # self.transform(), so doing it here will result in the transformations
            # being applied twice.
            #item.matrix = self.matrix * item.matrix
            item.viewport = self.viewport

            self.items.append(item)
            # Recursively append if elt is a <g> (group)
            if elt.tag == svg_ns + 'g':
                item.append(elt)

    def __repr__(self):
        return '<Group ' + self.id + " ({})".format( self.name ) + '>: ' + repr(self.items)

    def json(self):
        '''Return json formatted dictionary of group'''
        return {'Group ' + self.id + " ({})".format( self.name ) : self.items}

class Matrix:
    ''' SVG transformation matrix and its operations
    a SVG matrix is represented as a list of 6 values [a, b, c, d, e, f]
    (named vect hereafter) which represent the 3x3 matrix
    ((a, c, e)
     (b, d, f)
     (0, 0, 1))
    see http://www.w3.org/TR/SVG/coords.html#EstablishingANewUserSpace '''

    def __init__(self, vect=None):
        # Unit transformation vect by default
        if vect is None:
            vect = [1, 0, 0, 1, 0, 0]
        if len(vect) != 6:
            raise ValueError("Bad vect size %d" % len(vect))
        self.vect = list(vect)

    def __mul__(self, other):
        '''Matrix multiplication'''
        if isinstance(other, Matrix):
            a = self.vect[0] * other.vect[0] + self.vect[2] * other.vect[1]
            b = self.vect[1] * other.vect[0] + self.vect[3] * other.vect[1]
            c = self.vect[0] * other.vect[2] + self.vect[2] * other.vect[3]
            d = self.vect[1] * other.vect[2] + self.vect[3] * other.vect[3]
            e = self.vect[0] * other.vect[4] + self.vect[2] * other.vect[5] \
                    + self.vect[4]
            f = self.vect[1] * other.vect[4] + self.vect[3] * other.vect[5] \
                    + self.vect[5]
            return Matrix([a, b, c, d, e, f])

        if isinstance(other, Point):
            x = other.x * self.vect[0] + other.y * self.vect[2] + self.vect[4]
            y = other.x * self.vect[1] + other.y * self.vect[3] + self.vect[5]
            return Point(x,y)

        return NotImplemented

    def __str__(self):
        return str(self.vect)

    def xscale(self):
        '''Return the rotated x scalar value'''
        return self.vect[0]/abs(self.vect[0]) * math.sqrt(self.vect[0]**2 + self.vect[2]**2)
    def yscale(self):
        '''Return the rotated x scalar value'''
        return self.vect[3]/abs(self.vect[3]) * math.sqrt(self.vect[1]**2 + self.vect[3]**2)
    def rot(self):
        '''Return the angle of rotation from the matrix.

        https://math.stackexchange.com/questions/13150/extracting-rotation-scale-values-from-2d-transformation-matrix
        '''
        if self.vect[0] != 0:
            return Angle(math.atan2(-self.vect[2], self.vect[0]))
        if self.vect[3] != 0:
            return Angle(math.atan2(self.vect[1], self.vect[3]))
        return 0



class Path(Transformable):
    '''SVG <path> tag handler
    self.items contains all objects for path instructions.
    Calling .parse(...) will append new path instruction
    objects to items list.
    '''
    # class Path handles the <path> tag
    tag = 'path'
    COMMANDS = 'MmZzLlHhVvCcSsQqTtAa'

    def __init__(self, elt=None, *args, **kwargs):
        Transformable.__init__(self, elt, *args, **kwargs)
        if elt is not None:
            self.parse(elt.get('d'))

    def parse(self, pathstr:str):
        """Parse svg path string and build elements list"""

        pathlst = re.findall(number_re + r"|\ *[%s]\ *" % Path.COMMANDS, pathstr)

        pathlst.reverse()

        command = None
        current_pt = Point(0,0)
        start_pt = None

        while pathlst:
            if pathlst[-1].strip() in Path.COMMANDS:
                last_command = command
                command = pathlst.pop().strip()
                absolute = (command == command.upper())
                command = command.upper()
            else:
                if command is None:
                    raise ValueError("No command found at %d" % len(pathlst))

            if command == 'M':
            # MoveTo
                x = pathlst.pop()
                y = pathlst.pop()
                pt = Point(x, y)
                if absolute:
                    current_pt = pt
                else:
                    current_pt += pt
                start_pt = current_pt

                self.items.append(MoveTo(current_pt))

                # MoveTo with multiple coordinates means LineTo
                command = 'L'

            elif command == 'Z':
            # Close Path
                l = Segment(current_pt, start_pt)
                self.items.append(l)
                current_pt = start_pt

            elif command in 'LHV':
            # LineTo, Horizontal & Vertical line
                # extra coord for H,V
                if absolute:
                    x,y = current_pt.coord()
                else:
                    x,y = (0,0)

                if command in 'LH':
                    x = pathlst.pop()
                if command in 'LV':
                    y = pathlst.pop()

                pt = Point(x, y)
                if not absolute:
                    pt += current_pt

                self.items.append(Segment(current_pt, pt))
                current_pt = pt

            elif command in 'CQ':
                dimension = {'Q':3, 'C':4}
                bezier_pts = []
                bezier_pts.append(current_pt)
                for _ in range(1,dimension[command]):
                    x = pathlst.pop()
                    y = pathlst.pop()
                    pt = Point(x, y)
                    if not absolute:
                        pt += current_pt
                    bezier_pts.append(pt)

                self.items.append(Bezier(bezier_pts))
                current_pt = pt

            elif command in 'TS':
                # number of points to read
                nbpts = {'T':1, 'S':2}
                # the control point, from previous Bezier to mirror
                ctrlpt = {'T':1, 'S':2}
                # last command control
                last = {'T': 'QT', 'S':'CS'}

                bezier_pts = []
                bezier_pts.append(current_pt)

                if last_command in last[command]:
                    pt0 = self.items[-1].control_point(ctrlpt[command])
                else:
                    pt0 = current_pt
                pt1 = current_pt
                # Symmetrical of pt1 against pt0
                bezier_pts.append(pt1 + pt1 - pt0)

                for _ in range(0,nbpts[command]):
                    x = pathlst.pop()
                    y = pathlst.pop()
                    pt = Point(x, y)
                    if not absolute:
                        pt += current_pt
                    bezier_pts.append(pt)

                self.items.append(Bezier(bezier_pts))
                current_pt = pt

            elif command == 'A':
                rx = pathlst.pop()
                ry = pathlst.pop()
                xrot = pathlst.pop()
                # Arc flags are not necessarily separated numbers
                flags = pathlst.pop().strip()
                large_arc_flag = flags[0]
                if large_arc_flag not in '01':
                    logger.error("Arc parsing failure")
                    break

                if len(flags) > 1:  flags = flags[1:].strip()
                else:               flags = pathlst.pop().strip()
                sweep_flag = flags[0]
                if sweep_flag not in '01':
                    logger.error("Arc parsing failure")
                    break

                if len(flags) > 1:  x = flags[1:]
                else:               x = pathlst.pop()
                y = pathlst.pop()
                end_pt = Point(x, y)
                if not absolute: end_pt += current_pt
                self.items.append(
                    Arc(current_pt, rx, ry, xrot, large_arc_flag, sweep_flag, end_pt))
                current_pt = end_pt

            else:
                pathlst.pop()

    def __str__(self):
        return '\n'.join(str(x) for x in self.items)

    def __repr__(self):
        return '<Path ' + self.id + '>'

    def segments(self, precision=0) -> List[Segment]:
        '''Return a list of segments, each segment is ended by a MoveTo.
           A segment is a list of Points'''
        ret = []
        # group items separated by MoveTo
        for moveTo, group in itertools.groupby(self.items,
                lambda x: isinstance(x, MoveTo)):
            # Use only non MoveTo item
            if not moveTo:
                # Generate segments for each relevant item
                seg = [x.segments(precision) for x in group]
                # Merge all segments into one
                ret.append(list(itertools.chain.from_iterable(seg)))

        return ret

    def simplify(self, precision:float) -> List[Segment]:
        '''Simplify segment with precision:
           Remove any point which are ~aligned'''
        ret = []
        for seg in self.segments(precision):
            ret.append(simplify_segment(seg, precision))

        return ret

class Ellipse(Transformable):
    '''SVG <ellipse> tag handler
    An ellipse is created by the center point (center)
    the x radius (rx) and the y radius (ry).
    Setting these values will change the ellipse
    regardless if it was created by an xml element.

    If provided xml has a 'd' attribute or path
    then this will also parse that.
        (This is for support of inkscape arc objects)
    '''
    # class Ellipse handles the <ellipse> tag
    tag = 'ellipse'

    def __init__(self, elt=None, *args, **kwargs):
        Transformable.__init__(self, elt, *args, **kwargs)
        self.arc = False
        if elt is not None:
            self.center = Point(self.xlength(elt.get('cx')),
                                self.ylength(elt.get('cy')))
            self.rx = self.length(elt.get('rx'))
            self.ry = self.length(elt.get('ry'))
            if elt.get('d') is not None:
                self.arc = True
                self.path = Path(elt)
                self.path_str = elt.get('d')
        else:
            self.center = Point(0,0)
            self.rx = 0
            self.ry = 0

    def __repr__(self):
        return '<Ellipse ' + self.id + '>'

    def bbox(self) -> Tuple[Point, Point]:
        '''Bounding box'''
        #TODO change bounding box dependent on rotation
        pmin = self.center - Point(self.rx, self.ry)
        pmax = self.center + Point(self.rx, self.ry)
        return (pmin, pmax)

    def transform(self, matrix=None):
        '''Apply the provided matrix. Default (None)
        If no matrix is supplied then recursively apply
        it's already existing matrix to all items.
        Also apply to center, rx, and ry
        '''
        if matrix is None:
            matrix = self.matrix
        else:
            matrix *= self.matrix
        self.transform_styles(matrix)

        self.center = matrix * self.center
        self.rx = matrix.xscale()*self.rx
        self.ry = matrix.yscale()*self.ry
        self.rotation += math.degrees(matrix.rot().angle)
        self.matrix= matrix

    def P(self, t) -> Point:
        '''Return a Point on the Ellipse for t in [0..1] or % from angle 0 to the full circle.
        Rotation is not handled in this function.
        '''
        x = self.center.x + self.rx * math.cos(2 * math.pi * t)
        y = self.center.y + self.ry * math.sin(2 * math.pi * t)
        return Point(x,y)

    def segments(self, precision=0) -> List[Segment]:
        '''Flatten all curves to segments with target length of precision'''
        if self.arc:
            segs = self.path.segments(precision)
            return segs
        if max(self.rx, self.ry) < precision:
            return [[self.center]]

        p = [(0,self.P(0)), (1, self.P(1))]
        d = 2 * max(self.rx, self.ry)

        while d > precision:
            for (t1,_),(t2,_) in zip(p[:-1],p[1:]):
                t = t1 + (t2 - t1)/2.
                p.append((t, self.P(t)))
            p.sort(key=operator.itemgetter(0))
            d = Segment(p[0][1],p[1][1]).length()

        ret = [x.rot(math.radians(self.rotation), x=self.center.x, y=self.center.y) for __,x in p]
        return [ret]

    def simplify(self, __):
        '''Return self because a 3 point representation is already simple'''
        return self

# An arc is an ellipse with a beginning and an end point instead of an entire circumference
class Arc(Ellipse):
    '''This inherits from Ellipse but does not have a svg tag
    Because there are no arc tags this class converts the
    path data for an arc into an object that can be flattened.
    '''

    def __init__(self, start_pt, rx, ry, xrot, large_arc_flag, sweep_flag, end_pt):
        Ellipse.__init__(self, None)
        try:
            self.rx = float(rx)
            self.ry = float(ry)
            self.rotation = float(xrot)
            self.large_arc_flag = large_arc_flag=='1'
            self.sweep_flag = sweep_flag=='1'
        except:
            pass
        self.end_pts = [start_pt, end_pt]
        self.angles = []

        self.calcuate_center()

    def __repr__(self):
        return '<Arc ' + self.id + '>'

    def calcuate_center(self):
        '''Calculate the center point of the arc from the
        non-intuitively provided data in an svg path.

        This is done by creating rotated ellipses around
        the start and end point. Then choosing the correct
        intersection point based on the two arc choosing flags.
        If there is no intersection then the center is the midpoint
        between the beginning and end points.
        '''
        angle = Angle(math.radians(self.rotation))

        # set some variables that are used often to decrease size of final equations
        pts = self.end_pts
        cs2 = 2*angle.cos*angle.sin*(math.pow(self.ry, 2) - math.pow(self.rx, 2))
        rs = (math.pow(self.ry*angle.sin, 2) + math.pow(self.rx*angle.cos, 2))
        rc = (math.pow(self.ry*angle.cos, 2) + math.pow(self.rx*angle.sin, 2))


        # Create a line that passes through both intersection points
        y = -pts[0].x*(cs2) + pts[1].x*cs2 - 2*pts[0].y*rs + 2*pts[1].y*rs
        # Round to prevent floating point errors
        y = round(y, 10)
        # A vertical line will break the program so we cannot calculate with these equations
        if y != 0:
            # Finish calculating the line
            m = ( -2*pts[0].x*rc + 2*pts[1].x*rc - pts[0].y*cs2 + pts[1].y*cs2 ) / -y
            b = (
                math.pow(pts[0].x,2)*rc - math.pow(pts[1].x,2)*rc + pts[0].x*pts[0].y*cs2 -
                pts[1].x*pts[1].y*cs2 + math.pow(pts[0].y,2)*(rs) - math.pow(pts[1].y,2)*rs
            ) / -y

            # Now that we have a line we can setup a quadratic equation to solve for all intersection points
            qa = rc + m*cs2 + math.pow(m,2)*rs
            qb = -2*pts[0].x*rc + b*cs2 - pts[0].y*cs2 - m*pts[0].x*cs2 + 2*m*b*rs - 2*pts[0].y*m*rs
            qc = (
                math.pow(pts[0].x,2)*rc - b*pts[0].x*cs2 + pts[0].x*pts[0].y*cs2 + math.pow(b,2)*rs -
                2*b*pts[0].y*rs + math.pow(pts[0].y,2)*rs - math.pow(self.rx*self.ry, 2)
            )

        else:
            # When the slope is vertical we need to calculate with x instead of y
            x = (pts[0].x+pts[1].x)/2
            m=0
            b=x

            # The quadratic formula but solving for y instead of x and only when the slope is vertical
            qa = rs
            qb =  x*cs2 - pts[0].x*cs2 - 2*pts[0].y*rs
            qc = (
                math.pow(x,2)*rc - 2*x*pts[0].x*rc + math.pow(pts[0].x,2)*rc - x*pts[0].y*cs2 +
                pts[0].x*pts[0].y*cs2 + math.pow(pts[0].y,2)*rs - math.pow(self.rx*self.ry, 2)
            )

        # This is the value to see how many real solutions the quadratic equation has.
        # if root is negative then there are only imaginary solutions or no real solutions
        # if the root is 0 then there is one solution
        # otherwise there are two solutions
        root = math.pow(qb, 2) - 4*qa*qc

        # If there are no roots then we need to scale the arc to fit the points
        if root < 0:
            # Center point
            point = Point((pts[0].x + pts[1].x)/2,(pts[0].y + pts[1].y)/2)
            # Angle between center and one of the end points adjusted to remove rotation from original data
            ptAng = math.atan2(self.end_pts[0].y-point.y, self.end_pts[0].x-point.x) - angle.angle
            # Adjust the angle to compensate for ellipse irregularity
            ptAng = math.atan((self.rx/self.ry) * math.tan(ptAng))
            # Calculate scaling factor between provided ellipse and actual end points
            radius = math.sqrt(math.pow(self.rx*math.cos(ptAng),2) + math.pow(self.ry*math.sin(ptAng),2))
            dist = math.sqrt( math.pow(self.end_pts[0].x-point.x, 2)+math.pow(self.end_pts[0].y-point.y, 2))
            factor = dist/radius
            self.rx *= factor
            self.ry *= factor


        # finish solving the quadratic equation and find the corresponding points on the intersection line
        elif root == 0:
            xroot = (-qb+math.sqrt(root))/(2*qa)
            point = Point(xroot, xroot*m + b)
        # Using the provided large_arc and sweep flags to choose the correct root
        else:
            xroots = [(-qb+math.sqrt(root))/(2*qa), (-qb-math.sqrt(root))/(2*qa)]
            points = [Point(xroots[0], xroots[0]*m + b), Point(xroots[1], xroots[1]*m + b)]
            # Calculate the angle of the beginning point to the end point

            # If counterclockwise the two angles are the angle is within 180 degrees of each other:
            #   and no flags are set use the first center
            #   and the sweep flag is set use the second
            #   the large arc flag is set invert the previous selection

            # Don't save the angles because they are calculated from the first possible center.
            # This may change so we'll just recalculate the angles later on
            angles = []
            for pt in pts:
                pt = Point(pt.x-points[0].x, pt.y-points[0].y)
                pt.rot(math.radians(-self.rotation))
                pt = Point(pt.x/self.rx, pt.y/self.ry)
                angles.append(math.atan2(pt.y,pt.x)%(math.pi*2))
            target = 0
            if self.sweep_flag:
                target = 0 if (angles[0] - angles[1]) < 0 or (angles[0] - angles[1]) > math.pi else 1
            else:
                target = 1 if (angles[0] - angles[1]) < 0 or (angles[0] - angles[1]) > math.pi else 0

            point = points[target if not self.large_arc_flag else target ^ 1 ]


        # Swap the x and y results from when the intersection line is vertical because we solved for y instead of x
        # Also remove any insignificant floating point errors
        if y == 0:
            point = Point(round(point.y, 10), round(point.x, 10))
        else:
            point = Point(round(point.x, 10), round(point.y, 10))
        self.center = point

        # Calculate start and end angle of the un-rotated arc
        if len(self.angles) < 2:
            self.angles = []
            for pt in self.end_pts:
                pt = Point(pt.x-self.center.x, pt.y-self.center.y)
                pt = pt.rot(math.radians(-self.rotation))
                pt = Point(pt.x/self.rx, pt.y/self.ry)
                self.angles.append(math.atan2(pt.y,pt.x))

        if not self.sweep_flag and self.angles[0] < self.angles[1]:
            self.angles[0] += 2*math.pi
        elif self.sweep_flag and self.angles[1] < self.angles[0]:
            self.angles[1] += 2*math.pi

    def transform(self, matrix=None):
        super().transform(matrix)
        self.end_pts[0] = self.matrix * self.end_pts[0]
        self.end_pts[1] = self.matrix * self.end_pts[1]

    def segments(self, precision=0) -> List[Segment]:
        '''This returns segments as expected by the
        Path object. (A list of points. Not a list of lists of points)
        '''
        if max(self.rx, self.ry) < precision:
            return self.end_pts
        return Ellipse.segments(self, precision)[0]

    def P(self, t) -> Point:
        '''Return a Point on the Arc for t in [0..1] where t is the % from
        the start angle to the end angle.

        Final angle transformation is handled in Ellipse.segments
        '''
        x = self.center.x + self.rx * math.cos(((self.angles[1] - self.angles[0]) * t) + self.angles[0])
        y = self.center.y + self.ry * math.sin(((self.angles[1] - self.angles[0]) * t) + self.angles[0])
        return Point(x,y)



# A circle is a special type of ellipse where rx = ry = radius
class Circle(Ellipse):
    '''SVG <circle> tag handler
    This is an ellipse by rx and ry are equal.
    '''
    # class Circle handles the <circle> tag
    tag = 'circle'

    def __init__(self, elt=None, *args, **kwargs):
        if elt is not None:
            elt.set('rx', elt.get('r'))
            elt.set('ry', elt.get('r'))
        Ellipse.__init__(self, elt, *args, **kwargs)

    def __repr__(self):
        return '<Circle ' + self.id + '>'

class Rect(Path):
    '''SVG <rect> tag handler
    This decompiles a rectangle svg xml element into
    essentially a path with 4 segments.

    P1 and P2 are the opposing corner points.

    As of now corner radii are not supported.
    '''
    # class Rect handles the <rect> tag
    tag = 'rect'

    def __init__(self, elt=None, *args, **kwargs):
        Transformable.__init__(self, elt, *args, **kwargs)
        if elt is not None:
            p = Point(self.xlength(elt.get('x')),
                            self.ylength(elt.get('y')))
            width = self.xlength(elt.get("width"))
            height = self.xlength(elt.get("height"))

            rx = self.xlength(elt.get('rx'))
            ry = self.xlength(elt.get('ry'))
            if not rx: rx = ry if ry else 0
            if not ry: ry = rx if rx else 0
            if rx > width/2: rx = width/2
            if ry > height/2: ry = width/2
            if rx or ry:
                cmd = f'''M{p.x+rx} {p.y} a{rx} {ry} 0 0 0 {-rx} {ry} v{height-(ry*2)}
                a{rx} {ry} 0 0 0 {rx} {ry}   h{width-(rx*2)}
                a{rx} {ry} 0 0 0 {rx} {-ry}  v{-(height-(ry*2))}
                a{rx} {ry} 0 0 0 {-rx} {-ry} h{-(width-(rx*2))} z'''
            else:
                cmd = f'M{p.x},{p.y}v{height}h{width}v{-height}h{-width}'

            self.p = p
            self.width = width
            self.height = height
            self.rx = rx
            self.ry = ry

            self.parse(cmd)

    def __repr__(self):
        return '<Rect ' + self.id + '>'

class Line(Transformable):
    '''SVG <line> tag handler

    This is essentially a wrapper around the Segment class
    '''
    # class Line handles the <line> tag
    tag = 'line'

    def __init__(self, elt=None, *args, **kwargs):
        Transformable.__init__(self, elt, *args, **kwargs)
        if elt is not None:
            self.P1 = Point(self.xlength(elt.get('x1')),
                            self.ylength(elt.get('y1')))
            self.P2 = Point(self.xlength(elt.get('x2')),
                            self.ylength(elt.get('y2')))
            self.segment = Segment(self.P1, self.P2)

    def __repr__(self):
        return '<Line ' + self.id + '>'

    def bbox(self) -> Tuple[Point, Point]:
        '''Bounding box'''
        xmin = min([p.x for p in (self.P1, self.P2)])
        xmax = max([p.x for p in (self.P1, self.P2)])
        ymin = min([p.y for p in (self.P1, self.P2)])
        ymax = max([p.y for p in (self.P1, self.P2)])

        return (Point(xmin,ymin), Point(xmax,ymax))

    def transform(self, matrix=None):
        '''Apply the provided matrix. Default (None)
        If no matrix is supplied then recursively apply
        it's already existing matrix to all items.
        '''
        if matrix is None:
            matrix = self.matrix
        else:
            matrix *= self.matrix
        self.transform_styles(matrix)

        self.P1 = matrix * self.P1
        self.P2 = matrix * self.P2
        self.segment = Segment(self.P1, self.P2)

    def segments(self, __=0) -> List[Segment]:
        '''Return the segment of the line'''
        return [self.segment.segments()]


class Text(Transformable):
    '''SVG <text> tag handler
    Take provided xml text element and convert using ttf and otf fonts
    into path element that can be used.

    setting Text.default_font is important. If the listed font
    cannot be found this is the fall back value.

    A list of fonts installed on the system can be found by calling
        Text.load_system_fonts(...)
    this keeps all found font in memory after first time call to
    improve performance.

    All distinct text element, those that have different start locations
    or fonts, are stored in text in a list.

    Adding new strings can be done by calling add_text(...)
    and removing strings is done by removing the item from the text list

    Once all strings are properly configured in the text list running
    convert_to_path will append a list of path elements to the paths variable

    The bounding box will not report a valid size until convert_to_path has been ran.
    '''
    # class Text handles the <text> tag
    tag = 'text'

    default_font = None
    _system_fonts = {}
    _os_font_paths = {
        "Darwin": ["/Library/Fonts", "~/Library/Fonts"],
        "Linux": ["/usr/share/fonts","/usr/local/share/fonts","~/.local/share/fonts"],
        "Windows": ["C:/Windows/Fonts", "~/AppData/Local/Microsoft/Windows/Fonts"]
    }

    def __init__(self, elt=None, parent=None, *args, **kwargs):
        Transformable.__init__(self, elt, *args, **kwargs)

        self.bbox_points = [Point(0,0), Point(0,0)]
        self.paths = []

        if elt is not None:
            self.parse(elt, parent)
            if parent is None:
                self.convert_to_path(auto_transform=False)
        else:
            self.origin = Point(0,0)
            self.font_family = Text.default_font
            self.size = 12
            self.bold = "normal"
            self.italic = "normal"
            if self.font_family:
                self.font_file = self.find_font_file()
            self.text = []

    def set_font(self, font=None, bold=None, italic=None, size=None):
        '''Set the font of the current text element.
        font is expected to be a string of the font family name.
        bold is expected Boolean
        italic is expected Boolean
        size is expected int, but can work with string ending in px
        '''
        font = font if font else self.font_family
        bold = bold if bold else (self.bold.lower() != "normal")
        italic = italic if italic else (self.italic.lower() != "normal")
        size = size if size else self.size
        if isinstance(size, str):
            size = float(size.strip("px"))

        self.font_family = font
        self.size = size
        self.bold = "normal" if not bold else "bold"
        self.italic = "normal" if not italic else "italic"
        self.font_file = self.find_font_file()


    def add_text(self, text, origin=Point(0,0), inherit=True):
        '''Add text the list of text objects
        if the origin is not different then the parents origin or
        inherit is set to False then a new text element will
        be created an added to the strings tuple in the text list.
        '''
        if origin == self.origin and inherit:
            self.text.append((text, self))
        else:
            new_line = Text()
            new_line.set_font(
                font=self.font_family,
                bold=(self.bold != "normal"),
                italic=(self.italic != "normal"),
                size=self.size
            )

            new_line.origin = origin
            self.text.append((text, new_line))


    def parse(self, elt, parent):
        '''Read the useful data from the xml element.
        Since text tags can have nested text tags
        parse can be called multiple times for one text tag.
        However all nested tags should have parent set so
        they can inherit and append the proper values
        from their immediate parent
        '''
        x = elt.get('x')
        y = elt.get('y')

        # It seems that any values in style that override these values take precedence
        self.font_configs = {
            "font-family": elt.get('font-family'),
            "font-size": elt.get('font-size'),
            "font-weight": elt.get('font-weight'),
            "font-style": elt.get('font-style'),
        }
        for style in self.style:
            if style in self.font_configs.keys() and self.style[style]:
                self.font_configs[style] = self.style[style]

        if isinstance(self.font_configs["font-size"], str):
            self.font_configs["font-size"] = float(self.font_configs["font-size"].strip("px"))

        for config in self.font_configs:
            if self.font_configs[config] is None and parent is not None:
                self.font_configs[config] = parent.font_configs[config]

        self.font_family = self.font_configs["font-family"]
        self.size = self.font_configs["font-size"]
        self.bold = self.font_configs["font-weight"]
        self.italic = self.font_configs["font-style"]

        self.font_file = self.find_font_file()

        if parent is not None:
            x = parent.origin.x if x is None else float(x)
            y = parent.origin.y if y is None else float(y)
        x = 0 if x is None else float(x)
        y = 0 if y is None else float(y)
        self.origin = Point(x,y)

        self.text = [] if elt.text is None else [(elt.text, self)]
        for child in list(elt):
            Text(child, self)
        if parent is not None:
            parent.text.extend(self.text)
            if elt.tail is not None:
                parent.text.append((elt.tail, parent))

        del self.font_configs


    def find_font_file(self):
        '''This will look through the indexed fonts and
        attempt to find one with a matching font name and text style.

        -- Faux font styles are not supported ==

        If the styling cannot be found it will fall back to either
        italic or bold if both were asked for and there wasn't a style
        with both or regular if italic or bold are set but not found.

        If the target font cannot be found then the default is used if set and found.
        '''
        if self.font_family is None:
            if Text.default_font is None:
                global _font_warning_sent
                if not _font_warning_sent:
                    logger.error("Unable to find font because no font was specified.")
                    _font_warning_sent = True
                return None
            self.font_family = Text.default_font
        fonts = [fnt.strip().strip("'") for fnt in self.font_family.split(",")]
        if Text.default_font is not None: fonts.append(Text.default_font)

        font_files = None
        target_font = None
        for fnt in fonts:
            if Text.load_system_fonts().get(fnt) is not None:
                target_font = fnt
                font_files = Text.load_system_fonts().get(fnt)
                break
        if font_files is None:
            # We are unable to find a font and since there is no default font stop building font data
            logger.error("Unable to find font(s) \"{}\"{}".format(
                self.font_family,
                " and no default font specified" if Text.default_font is None else f" or default font \"{Text.default_font}\""
            ))
            self.paths = []
            return

        bold = self.bold is not None and self.bold.lower() != "normal"
        italic = self.italic is not None and self.italic.lower() != "normal"

        reg = ["Regular", "Book"]
        bol = ["Bold", "Demibold"]
        ita = ["Italic", "Oblique"]

        search = reg
        if bold and not italic:
            search = bol
        elif italic and not bold:
            search = ita
        elif italic and bold:
            search = [f"{b} {i}" if n == 0 else f"{i} {b}" for b in bol for i in ita for n in range(2)]
        tar_font = list(filter(None, [font_files.get(style) for style in search]))
        if len(tar_font) == 0 and len(font_files.keys()) == 1:
            tar_font = [font_files[list(font_files.keys())[0]]]
            logger.warning("Font \"{}\" does not natively support style \"{}\" using \"{}\" instead".format(
                target_font, search[0], list(font_files.keys())[0]))
        elif len(tar_font) == 0 and italic and bold:
            orig_search = search[0]
            search = []
            search.extend(ita)
            search.extend(bol)
            search.extend(reg)
            search.extend(list(font_files.keys()))
            for style in search:
                if font_files.get(style) is not None:
                    tar_font = [font_files[style]]
                    logger.warning("Font \"{}\" does not natively support style \"{}\" using \"{}\" instead".format(
                        target_font, orig_search, style))
                    break
        return tar_font[0]


    def convert_to_path(self, auto_transform=True):
        ''' Read the vector data from the ttf/otf file and
        convert it into a path string for each letter and
        parse the path string by a Path instance.

        if auto_transform is True then this calls self.transform()
        at the end to apply all transformations on the paths.

        This should only be called once so double check transform()
        is never called elsewhere.
        '''
        self.paths = []
        if not self.text: return
        prev_origin = self.text[0][1].origin

        offset = Point(prev_origin.x, prev_origin.y)
        for text, attrib in self.text:

            if attrib.font_file is None or attrib.font_family is None:
                continue
            size = attrib.size
            ttf = ttFont.TTFont(attrib.font_file)
            offset.y = attrib.origin.y + ttf["head"].unitsPerEm
            scale = size/offset.y

            if prev_origin != attrib.origin:
                prev_origin = attrib.origin
                offset.x = attrib.origin.x

            path = []
            for char in text:

                pathbuf = ""
                try: glf = ttf.getGlyphSet()[ttf.getBestCmap()[ord(char)]]
                except KeyError:
                    logger.warning('Unsuported character in <text> element "{}"'.format(char))
                    #txt = txt.replace(char, "")
                    continue

                pen = SVGPathPen(ttf.getGlyphSet)
                glf.draw(pen)

                for cmd in pen._commands:
                    pathbuf += cmd + ' '

                if len(pathbuf) > 0:
                    path.append(Path())
                    path[-1].parse(pathbuf)
                    # Apply the scaling then the translation
                    translate = Matrix([1,0,0,-1,offset.x,size+attrib.origin.y]) * Matrix([scale,0,0,scale,0,0])
                    # This queues the translations until .transform() is called
                    path[-1].matrix =  translate * path[-1].matrix

                offset.x += (scale*glf.width)

            self.paths.append(path)
        if auto_transform:
            self.transform()

    def bbox(self) -> Tuple[Point, Point]:
        '''Find the bounding box of all the paths that make
        each letter.
        This will only work if there are available paths.
        '''
        if self.paths is None or len(self.paths) == 0:
            return [Point(0,0),Point(0,0)]

        bboxes = [path.bbox() for paths in self.paths for path in paths]

        return (
            Point(min(bboxes, key=lambda v: v[0].x)[0].x, min(bboxes, key=lambda v: v[0].y)[0].y),
            Point(max(bboxes, key=lambda v: v[1].x)[1].x, max(bboxes, key=lambda v: v[1].y)[1].y),
        )

    def transform(self, matrix=None):
        '''Apply the provided matrix. Default (None)
        If no matrix is supplied then recursively apply
        it's already existing matrix to all items.
        '''
        if matrix is None:
            matrix = self.matrix
        else:
            matrix *= self.matrix
        self.transform_styles(matrix)

        self.origin = matrix * self.origin
        for paths in self.paths:
            for path in paths:
                path.transform(matrix)

    def segments(self, precision=0) -> List[Segment]:
        '''Get a list of all points in all paths
        with provide precision.
        This will only work if there are available paths.
        '''
        segs = []
        for paths in self.paths:
            for path in paths:
                segs.extend(path.segments(precision))
        return segs

    @staticmethod
    def load_system_fonts(reload:bool=False) -> List[dict]:
        '''Find all fonts in common locations on the file system
        To properly read all fonts they need to be parsed so this
        is inherently slow on systems with many fonts.
        To prevent long parsing time all the results are cached
        and the cached results are returned next time this function
        is called.
        If a force reload of all indexed fonts is desirable setting
        reload to True will clear the cache and re-index the system.
        '''
        if reload:
            Text._system_fonts = {}
        if len(Text._system_fonts.keys()) < 1:
            fonts_files = []
            logger.info("Loading system fonts.")
            for path in Text._os_font_paths[platform.system()]:
                try:
                    fonts_files.extend([os.path.join(dp, f) for dp, dn, fn in os.walk(os.path.expanduser(path)) for f in fn])
                except:
                    pass

            for ffile in fonts_files:
                try:
                    font = ttFont.TTFont(ffile)
                    name = font["name"].getName(1,1,0).toStr()
                    style = font["name"].getName(2,1,0).toStr()
                    if Text._system_fonts.get(name) is None:
                        Text._system_fonts[name] = {style:ffile}
                    elif Text._system_fonts[name].get(style) is None:
                        Text._system_fonts[name][style] = ffile
                except:
                    pass
            logger.debug(f"  Found {len(Text._system_fonts.keys())} fonts in system")
        return Text._system_fonts


class JSONEncoder(json.JSONEncoder):
    ''' overwrite JSONEncoder for svg classes which have defined a .json() method '''
    def default(self, obj):
        ''' overwrite default function to handle svg classes '''
        if not isinstance(obj, tuple(svgClass.values() + [Svg])):
            return json.JSONEncoder.default(self, obj)

        if not hasattr(obj, 'json'):
            return repr(obj)

        return obj.json()

## Code executed on module load ##

# Make fontTools more quiet
loggingTools.configLogger(level=logging.INFO)


# SVG tag handler classes are initialized here
# (classes must be defined before)

svgClass = {}
# Register all classes with attribute 'tag' in svgClass dictionary
for name, cls in inspect.getmembers(sys.modules[__name__], inspect.isclass):
    tag = getattr(cls, 'tag', None)
    if tag:
        svgClass[svg_ns + tag] = cls




# Copyright (C) 2022 -- svg2mod developers < GitHub.com / svg2mod >

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

'''
This module contains the necessary tools to convert from
the svg objects provided from the svg2mod.svg module to
KiCad file formats.
This currently supports both the pretty format and
the legacy mod format.
'''


# Copyright (C) 2022 -- svg2mod developers < GitHub.com / svg2mod >

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

'''
The command line interface for using svg2mod
as a tool in a terminal.
'''

import argparse
import logging
import os
import shlex
import sys
import traceback

import svg2mod.coloredlogger as coloredlogger
from svg2mod.coloredlogger import logger, unfiltered_logger
from svg2mod import svg
from svg2mod.exporter import (DEFAULT_DPI, Svg2ModExportLatest,
                              Svg2ModExportLegacy, Svg2ModExportLegacyUpdater,
                              Svg2ModExportPretty)
from svg2mod.importer import Svg2ModImport

#----------------------------------------------------------------------------

def main():
    '''This function handles the scripting package calls.
    It is setup to read the arguments from `get_arguments()`
    then parse the target svg file and output all converted
    objects into a kicad footprint module.
    '''

    args,_ = get_arguments()


    # Setup root logger to use terminal colored outputs as well as stdout and stderr
    coloredlogger.split_logger(logger)

    if args.debug_print:
        logger.setLevel(logging.DEBUG)
    elif args.verbose_print:
        logger.setLevel(logging.INFO)
    else:
        logger.setLevel(logging.WARNING)

    if args.list_fonts:
        fonts = svg.Text.load_system_fonts()
        unfiltered_logger.info("Font Name: list of supported styles.")
        for font in fonts:
            fnt_text = f"  {font}:"
            for styles in fonts[font]:
                fnt_text += f" {styles},"
            fnt_text = fnt_text.strip(",")
            unfiltered_logger.info(fnt_text)
        sys.exit(0)
    if args.default_font:
        svg.Text.default_font = args.default_font

    pretty = args.format in ['pretty','latest']
    use_mm = args.units == 'mm'

    if pretty:

        if not use_mm:
            logger.critical("Error: decimal units only allowed with legacy output type")
            sys.exit( -1 )

    try:
        # Import the SVG:
        imported = Svg2ModImport(
            args.input_file_name,
            args.module_name,
            args.module_value,
            args.ignore_hidden,
            args.force_layer
        )

        # Pick an output file name if none was provided:
        if args.output_file_name is None:

            args.output_file_name = os.path.splitext(
                os.path.basename( args.input_file_name )
            )[ 0 ]

        # Append the correct file name extension if needed:
        if pretty:
            extension = ".kicad_mod"
        else:
            extension = ".mod"
        if args.output_file_name[ - len( extension ) : ] != extension:
            args.output_file_name += extension

        # Create an exporter:
        if pretty:
            exported = (Svg2ModExportPretty if args.format == "pretty" else Svg2ModExportLatest)(
                imported,
                args.output_file_name,
                args.center,
                args.scale_factor,
                args.precision,
                dpi = args.dpi,
                pads = args.convert_to_pads,
            )

        else:

            # If the module file exists, try to read it:
            exported = None
            if os.path.isfile( args.output_file_name ):

                try:
                    exported = Svg2ModExportLegacyUpdater(
                        imported,
                        args.output_file_name,
                        args.center,
                        args.scale_factor,
                        args.precision,
                        args.dpi,
                    )

                except Exception as e:
                    raise e

            # Write the module file:
            if exported is None:
                exported = Svg2ModExportLegacy(
                    imported,
                    args.output_file_name,
                    args.center,
                    args.scale_factor,
                    args.precision,
                    use_mm = use_mm,
                    dpi = args.dpi,
                )

        cmd_args = [os.path.basename(sys.argv[0])] + sys.argv[1:]
        cmdline = ' '.join(shlex.quote(x) for x in cmd_args)

        # Export the footprint:
        exported.write(cmdline)
    except Exception as e:
        if args.debug_print:
            traceback.print_exc()
        else:
            logger.critical(f'Unhandled exception (Exiting)\n {type(e).__name__}: {e} ')
        exit(-1)

#----------------------------------------------------------------------------

def get_arguments():
    ''' Return an instance of pythons argument parser
    with all the command line functionalities arguments
    '''

    parser = argparse.ArgumentParser(
        description = (
            'Convert Inkscape SVG drawings to KiCad footprint modules.'
        )
    )

    mux = parser.add_mutually_exclusive_group(required=True)

    mux.add_argument(
        '-i', '--input-file',
        type = str,
        dest = 'input_file_name',
        metavar = 'FILENAME',
        help = "Name of the SVG file",
    )

    parser.add_argument(
        '-o', '--output-file',
        type = str,
        dest = 'output_file_name',
        metavar = 'FILENAME',
        help = "Name of the module file",
    )

    parser.add_argument(
        '-c', '--center',
        dest = 'center',
        action = 'store_const',
        const = True,
        help = "Center the module to the center of the bounding box",
        default = False,
    )

    parser.add_argument(
        '-P', '--convert-pads',
        dest = 'convert_to_pads',
        action = 'store_const',
        const = True,
        help = "Convert any artwork on Cu layers to pads",
        default = False,
    )

    parser.add_argument(
        '-v', '--verbose',
        dest = 'verbose_print',
        action = 'store_const',
        const = True,
        help = "Print more verbose messages",
        default = False,
    )

    parser.add_argument(
        '--debug',
        dest = 'debug_print',
        action = 'store_const',
        const = True,
        help = "Print debug level messages",
        default = False,
    )

    parser.add_argument(
        '-x', '--exclude-hidden',
        dest = 'ignore_hidden',
        action = 'store_const',
        const = True,
        help = "Do not export hidden objects",
        default = False,
    )

    parser.add_argument(
        '--force', '--force-layer',
        type = str,
        dest = 'force_layer',
        metavar = 'LAYER',
        help = "Force everything into the single provided layer",
        default = None,
    )

    parser.add_argument(
        '-d', '--dpi',
        type = int,
        dest = 'dpi',
        metavar = 'DPI',
        help = "DPI of the SVG file (int)",
        default = DEFAULT_DPI,
    )

    parser.add_argument(
        '-f', '--factor',
        type = float,
        dest = 'scale_factor',
        metavar = 'FACTOR',
        help = "Scale paths by this factor",
        default = 1.0,
    )

    parser.add_argument(
        '-p', '--precision',
        type = float,
        dest = 'precision',
        metavar = 'PRECISION',
        help = "Smoothness for approximating curves with line segments. Input is the approximate length for each line segment in SVG pixels (float)",
        default = 5.0,
    )
    parser.add_argument(
        '--format',
        type = str,
        dest = 'format',
        metavar = 'FORMAT',
        choices = [ 'legacy', 'pretty', 'latest'],
        help = "Output module file format (legacy|pretty|latest). 'latest' introduces features used in kicad >= 6",
        default = 'latest',
    )

    parser.add_argument(
        '--name', '--module-name',
        type = str,
        dest = 'module_name',
        metavar = 'NAME',
        help = "Base name of the module",
        default = "svg2mod",
    )

    parser.add_argument(
        '--units',
        type = str,
        dest = 'units',
        metavar = 'UNITS',
        choices = [ 'decimal', 'mm' ],
        help = "Output units, if output format is legacy (decimal|mm)",
        default = 'mm',
    )

    parser.add_argument(
        '--value', '--module-value',
        type = str,
        dest = 'module_value',
        metavar = 'VALUE',
        help = "Value of the module",
        default = "G***",
    )

    parser.add_argument(
        '-F', '--default-font',
        type = str,
        dest = 'default_font',
        help = "Default font to use if the target font in a text element cannot be found",
    )

    mux.add_argument(
        '-l', '--list-fonts',
        dest = 'list_fonts',
        const = True,
        default = False,
        action = "store_const",
        help = "List all fonts that can be found in common locations",
    )

    return parser.parse_args(), parser

#----------------------------------------------------------------------------

if __name__ == "__main__":
    main()

#----------------------------------------------------------------------------



# Copyright (C) 2022 -- svg2mod developers < GitHub.com / svg2mod >

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
'''
A simple modification to the formatter class in the python logger to allow
ANSI color codes based on the logged message's level
'''

import logging
import sys

#----------------------------------------------------------------------------

#Setup and configure svg2mod and svg2mod-unfiltered loggers

logger = logging.getLogger("svg2mod")
unfiltered_logger = logging.getLogger("svg2mod-unfiltered")
_sh = logging.StreamHandler(sys.stdout)

logger.addHandler(_sh)
unfiltered_logger.addHandler(_sh)

logger.setLevel(logging.DEBUG)

# Add a second logger that will bypass the log level and output anyway
# It is a good practice to send only messages level INFO via this logger
unfiltered_logger.setLevel(logging.INFO)

# This can be used sparingly as follows:
#---------
# unfiltered_logger.info("Message Here")
#---------


#----------------------------------------------------------------------------

class Formatter(logging.Formatter):
    '''Extend formatter to add colored output functionality '''

    # ASCII escape codes for supporting terminals
    color = {
        logging.CRITICAL: "\033[91m\033[7m", #Set red and swap background and foreground
        logging.ERROR: "\033[91m", #Set red
        logging.WARNING: "\033[93m", #Set yellow
        logging.DEBUG: "\033[90m", #Set dark gray/black
        logging.INFO: "" #Do nothing
    }
    reset = "\033[0m" # Reset the terminal back to default color/emphasis

    #------------------------------------------------------------------------

    def __init__(self, fmt="%(message)s", datefmt=None, style="%"):
        super().__init__(fmt, datefmt, style)

    #------------------------------------------------------------------------

    def format(self, record):
        '''Overwrite the format function.
        This saves the original style, overwrites it to support
        color, sends the message to the super.format, and
        finally returns the style to the original format
        '''
        if sys.stdout.isatty():
            fmt_org = self._style._fmt
            self._style._fmt = Formatter.color[record.levelno] + fmt_org + Formatter.reset
        result = logging.Formatter.format(self, record)
        if sys.stdout.isatty():
            self._style._fmt = fmt_org
        return result

    #------------------------------------------------------------------------

#----------------------------------------------------------------------------

def split_logger(logger, formatter=Formatter(), brkpoint=logging.WARNING):
    '''This will split logging messages at the specified break point. Anything higher
    will be sent to sys.stderr and everything else to sys.stdout
    '''
    for hndl in logger.handlers:
        logger.removeHandler(hndl)

    hdlrerr = logging.StreamHandler(sys.stderr)
    hdlrerr.addFilter(lambda msg: brkpoint <= msg.levelno)

    hdlrout = logging.StreamHandler(sys.stdout)
    hdlrout.addFilter(lambda msg: brkpoint > msg.levelno)

    hdlrerr.setFormatter(formatter)
    hdlrout.setFormatter(formatter)
    logger.addHandler(hdlrerr)
    logger.addHandler(hdlrout)

#----------------------------------------------------------------------------


#!/usr/bin/env python
# -*- coding: utf-8 -*-
'''
to install system wide use `pip install .`
testing is done via `python setup.py test`
'''

import subprocess
import setuptools


try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


with open('README.md') as readme_file:
    readme = readme_file.read()

tag = ""

try:
    ps = subprocess.check_output(["git","describe","--tag"], stderr=subprocess.STDOUT)
    tag = ps.decode('utf-8').strip()
    tag = tag.replace("-", ".dev", 1).replace("-", "+")
except:
    tag = "0.dev0"

requirements = [
    "fonttools"
]

setup_requirements = [
	"pytest-runner", "pytest-pylint",
]
test_requirements = [
	"pytest", "pylint", "pyenchant",
]

setup(
    name='svg2mod',
    version=tag,
    description="Convert an SVG file to a KiCad footprint.",
    long_description_content_type='text/markdown',
    long_description=readme,
    author='https://github.com/svg2mod',
    author_email='',
    url='https://github.com/svg2mod/svg2mod',
    packages=setuptools.find_packages(),
    entry_points={'console_scripts':['svg2mod = svg2mod.cli:main']},
    package_dir={'svg2mod':'svg2mod'},
    include_package_data=True,
    package_data={'kipart': ['*.gif', '*.png']},
    scripts=[],
    install_requires=requirements,
    license="GPLv2+",
    zip_safe=False,
    keywords='svg2mod, KiCAD, inkscape',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        'Intended Audience :: Manufacturing',
        'License :: OSI Approved :: GNU General Public License v2 or later (GPLv2+)',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
    ],
	setup_requires=setup_requirements,
    test_suite='test',
    tests_require=test_requirements
)



'''
A SVG parser with tools to convert an XML svg file
to objects that can be simplified into points.
'''
#__all__ = ['geometry', 'svg']

from .svg import *

def parse(filename):
    '''Take in a filename and return a SVG object of parsed file'''
    return Svg(filename)
